# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from dataclasses import dataclass
from typing import List, Optional

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from crossfit.backend.torch.hf.model import HFModel
from transformers import AutoConfig, AutoTokenizer

from nemo_curator.classifiers.base import (
    DistributedDataClassifier,
    HFDeberta,
    _get_suggest_memory_for_classifier,
    _run_classifier_helper,
)
from nemo_curator.datasets import DocumentDataset

DOMAIN_IDENTIFIER = "nvidia/domain-classifier"
DOMAIN_BASE_MODEL = "microsoft/deberta-v3-base"
MULTILINGUAL_DOMAIN_IDENTIFIER = "nvidia/multilingual-domain-classifier"
MULTILINGUAL_DOMAIN_BASE_MODEL = "microsoft/mdeberta-v3-base"


@dataclass
class DomainModelConfig:
    identifier: str = DOMAIN_IDENTIFIER
    base_model: str = DOMAIN_BASE_MODEL
    fc_dropout: float = 0.2
    max_len: int = 512


class DomainModel(HFModel):
    def __init__(
        self,
        config: DomainModelConfig,
        autocast: bool = False,
        max_mem_gb: Optional[int] = None,
    ):
        self.config = config
        self.autocast = autocast
        if max_mem_gb is None:
            max_mem_gb = _get_suggest_memory_for_classifier()

        super().__init__(self.config.base_model, max_mem_gb=max_mem_gb)

    def load_model(self, device: str = "cuda"):
        model = HFDeberta.from_pretrained(self.config.identifier)
        model.set_autocast(self.autocast)
        model = model.to(device)
        return model.eval()

    def load_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.config.identifier)

    def load_config(self):
        return AutoConfig.from_pretrained(self.config.identifier)


class _DomainClassifier(DistributedDataClassifier):
    """
    Parent class for DomainClassifier and MultilingualDomainClassifier,
    since their implementations are almost identical.
    """

    def __init__(
        self,
        multilingual: bool = False,
        filter_by: Optional[List[str]] = None,
        batch_size: int = 256,
        text_field: str = "text",
        pred_column: str = "domain_pred",
        prob_column: Optional[str] = None,
        max_chars: int = 2000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        self.multilingual = multilingual

        if multilingual:
            config = AutoConfig.from_pretrained(MULTILINGUAL_DOMAIN_IDENTIFIER)
            model_config = DomainModelConfig(
                identifier=MULTILINGUAL_DOMAIN_IDENTIFIER,
                base_model=MULTILINGUAL_DOMAIN_BASE_MODEL,
            )
        else:
            config = AutoConfig.from_pretrained(DOMAIN_IDENTIFIER)
            model_config = DomainModelConfig(
                identifier=DOMAIN_IDENTIFIER,
                base_model=DOMAIN_BASE_MODEL,
            )

        self.text_field = text_field
        self.prob_column = prob_column
        self.labels = list(config.label2id.keys())
        self.labels.sort(key=lambda x: config.label2id[x])
        self.out_dim = len(self.labels)

        model = DomainModel(
            config=model_config, autocast=autocast, max_mem_gb=max_mem_gb
        )

        super().__init__(
            model=model,
            labels=self.labels,
            filter_by=filter_by,
            batch_size=batch_size,
            out_dim=self.out_dim,
            pred_column=pred_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
        )

    def _run_classifier(self, dataset: DocumentDataset) -> DocumentDataset:
        if self.multilingual:
            print("Starting multilingual domain classifier inference", flush=True)
        else:
            print("Starting domain classifier inference", flush=True)

        df = dataset.df
        df = _run_classifier_helper(
            df=df,
            model=self.model,
            labels=self.labels,
            max_chars=self.max_chars,
            batch_size=self.batch_size,
            label_col=self.pred_column,
            text_field=self.text_field,
            prob_col=self.prob_column,
        )
        return DocumentDataset(df)


class DomainClassifier(_DomainClassifier):
    """
    DomainClassifier is a specialized classifier designed for English text domain classification tasks,
    utilizing the NemoCurator Domain Classifier (https://huggingface.co/nvidia/domain-classifier) model.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large datasets.

    Attributes:
        filter_by (list[str], optional): The classes to filter the dataset by.
                                         If None, all classes will be included. Defaults to None.
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The field in the dataset that should be classified.
        pred_column (str): The column name where predictions will be stored. Defaults to "domain_pred".
        prob_column (str, optional): The column name where prediction probabilities will be stored. Defaults to None.
        max_chars (int): The maximum number of characters in each document to consider for classification. Defaults to 2000.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        filter_by: Optional[List[str]] = None,
        batch_size: int = 256,
        text_field: str = "text",
        pred_column: str = "domain_pred",
        prob_column: Optional[str] = None,
        max_chars: int = 2000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        super().__init__(
            multilingual=False,
            filter_by=filter_by,
            batch_size=batch_size,
            text_field=text_field,
            pred_column=pred_column,
            prob_column=prob_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )


class MultilingualDomainClassifier(_DomainClassifier):
    """
    MultilingualDomainClassifier is a specialized classifier designed for domain classification tasks,
    utilizing the NemoCurator Multilingual Domain Classifier (https://huggingface.co/nvidia/multilingual-domain-classifier) model.
    It supports domain classification across 52 languages.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large datasets.

    Attributes:
        filter_by (list[str], optional): The classes to filter the dataset by.
                                         If None, all classes will be included. Defaults to None.
        batch_size (int): The number of samples per batch for inference. Defaults to 256.
        text_field (str): The field in the dataset that should be classified.
        pred_column (str): The column name where predictions will be stored. Defaults to "domain_pred".
        prob_column (str, optional): The column name where prediction probabilities will be stored. Defaults to None.
        max_chars (int): The maximum number of characters in each document to consider for classification. Defaults to 2000.
        device_type (str): The type of device to use for inference, either "cuda" or "cpu". Defaults to "cuda".
        autocast (bool): Whether to use mixed precision for faster inference. Defaults to True.
        max_mem_gb (int, optional): The maximum amount of memory in GB to allocate for the model. If None,
                                      it defaults to the available GPU memory minus 4 GB.

    """

    def __init__(
        self,
        filter_by: Optional[List[str]] = None,
        batch_size: int = 256,
        text_field: str = "text",
        pred_column: str = "domain_pred",
        prob_column: Optional[str] = None,
        max_chars: int = 2000,
        device_type: str = "cuda",
        autocast: bool = True,
        max_mem_gb: Optional[int] = None,
    ):
        super().__init__(
            multilingual=True,
            filter_by=filter_by,
            batch_size=batch_size,
            text_field=text_field,
            pred_column=pred_column,
            prob_column=prob_column,
            max_chars=max_chars,
            device_type=device_type,
            autocast=autocast,
            max_mem_gb=max_mem_gb,
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dask

from .modules import *
from .package_info import (
    __contact_emails__,
    __contact_names__,
    __description__,
    __download_url__,
    __homepage__,
    __keywords__,
    __license__,
    __package_name__,
    __repository_url__,
    __shortversion__,
    __version__,
)
from .services import (
    AsyncLLMClient,
    AsyncOpenAIClient,
    LLMClient,
    NemoDeployClient,
    OpenAIClient,
)
from .utils.distributed_utils import get_client, get_network_interfaces

# Dask will automatically convert the list score type
# to a string without this option.
# See https://github.com/NVIDIA/NeMo-Curator/issues/33
# This also happens when reading and writing to files
dask.config.set({"dataframe.convert-string": False})

__all__ = [
    "__contact_emails__",
    "__contact_names__",
    "__description__",
    "__download_url__",
    "__homepage__",
    "__keywords__",
    "__license__",
    "__package_name__",
    "__repository_url__",
    "__shortversion__",
    "__version__",
]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import time
import warnings
from itertools import pairwise
from typing import Optional, Union

import cudf
import dask_cudf
import numpy as np
import pandas as pd
import pyarrow as pa

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix


class BucketsToEdges:
    """
    Maps buckets generated from LSH into an edgelist that
    can be processed further by Connected Components to find duplicate
    documents
    """

    def __init__(
        self,
        cache_dir: str = None,
        id_fields: Union[list, str] = "id",
        str_id_name: str = "id",
        bucket_field: str = "_bucket_id",
        logger: Union[logging.LoggerAdapter, str] = "./",
        profile_dir: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        cache_dir: str or None
          If specified, will compute & write the edgelist to a file
        id_fields: list or str
          id fields of documents in buckets_df
        str_id_name: str
          Ignored if there is a single id field. Multiple id fields
          will be combined into a single id field with the given name.
        bucket_field: str
          Column denoting bucket ID
        num_buckets: Number of bands/buckets to create from the minhash signature.
          Hashes_per_signature = num_hashes / num_buckets
        """
        self.cache_dir = cache_dir
        self.id_fields = [id_fields] if isinstance(id_fields, str) else id_fields
        self.str_id_name = str_id_name if len(self.id_fields) > 1 else self.id_fields[0]
        self.output_ids = [f"{self.str_id_name}_x", f"{self.str_id_name}_y"]
        self.bucket_field = bucket_field
        self.profile_dir = profile_dir
        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "Buckets_to_Edges.log"),
                name="Buckets_to_Edges",
            )
        else:
            self._logger = logger

    @staticmethod
    def _combine_multiple_ids(
        input_df: cudf.DataFrame, input_id_fields: list, output_id_field: str
    ) -> cudf.DataFrame:
        if output_id_field in input_df.columns:
            raise ValueError(
                f"Input df already contains column named: {output_id_field}"
            )

        output_df = input_df.copy()[input_df.columns.difference(input_id_fields)]

        output_df[output_id_field] = input_df[input_id_fields[0]].astype(str)
        for input_field in input_id_fields[1:]:
            output_df[output_id_field] = output_df[output_id_field] = (
                input_df[input_id_fields[0]].astype(str)
                + "-"
                + input_df[input_field].astype(str)
            )

        return output_df

    def buckets_to_edges(
        self,
        buckets_df: cudf.DataFrame,
    ) -> cudf.DataFrame:

        grouped_buckets = (
            buckets_df.groupby(self.bucket_field)[self.str_id_name]
            .agg(list)
            .list.sort_values()
        )
        bucket_docs = grouped_buckets.to_arrow().to_pylist()
        edges = []
        # Create pairs of all documents within a bucket since they are near duplicates
        # Effectively create a edge list of all near duplicate documents
        for bucket_doc in bucket_docs:
            edges.extend(pairwise(bucket_doc))
        edges = pd.DataFrame(edges, columns=self.output_ids)
        edges = pa.Table.from_pandas(edges)
        result_df = cudf.DataFrame.from_arrow(edges)
        del edges
        result_df = result_df.drop_duplicates(self.output_ids).reset_index(drop=True)
        result_df["jaccard"] = np.float32(1.0)
        return result_df

    def __call__(self, dataset: DocumentDataset) -> DocumentDataset:
        buckets_df = dataset.df
        self._logger.info(f"Starting conversion of LSH Buckets to Graph Edgelist")
        if len(self.id_fields) > 1:
            buckets_df = buckets_df.map_partitions(
                BucketsToEdges._combine_multiple_ids,
                input_id_fields=self.id_fields,
                output_id_field=self.str_id_name,
            )

        meta = [(output_id, str) for output_id in self.output_ids]
        meta.append(("jaccard", np.float32))
        edges_df = buckets_df.map_partitions(self.buckets_to_edges, meta=meta)

        if self.cache_dir is None:
            return DocumentDataset(edges_df)

        write_path = os.path.join(self.cache_dir, "_edges.parquet")
        if os.path.exists(write_path):
            warnings.warn(
                f"Output path {write_path} already exists and will be overwritten"
            )
        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir,
            "bucket-to-edges",
        ):
            edges_df.to_parquet(write_path, write_index=False, overwrite=True)
        self._logger.info(
            f"Time taken for Converted Buckets To Edgelist = {time.time() - t0}s and output written at {write_path}"
        )

        return DocumentDataset(
            dask_cudf.read_parquet(write_path, split_row_groups=False)
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Union

from nemo_curator.modifiers import DocumentModifier


class Slicer(DocumentModifier):
    """
    Slices a document based on indices or strings.
    """

    def __init__(
        self,
        left: Optional[Union[int, str]] = 0,
        right: Optional[Union[int, str]] = None,
        include_left: bool = True,
        include_right: bool = True,
        strip: bool = True,
    ):
        """
        Args:
            left (Union[int, str], optional): If the provided value is an int, slice the string from this index (inclusive).
                If the provided value is a str, slice the string from the first occurence of this substring.
            right (Union[int, str], optional): If the provided value is an int, slice the string to this index (exclusive).
                If the provided value is a str, slice the string to the last occurence of this substring. If None,
                right is set to the length of the string.
            include_left (bool): Only used if `left` is a string. If True, the value of `left` is included in the
                slicing result. Defaults to False.
            include_right (bool): Only used if `right` is a string. If True, the value of `right` is included in the
                slicing result. Defaults to False.
            strip (bool): If True, strip the resulting string.
        """
        super().__init__()
        self._left = left
        self._right = right
        self._include_left = include_left
        self._include_right = include_right
        self._strip = strip

    def modify_document(self, text: str) -> str:
        # Determine start index based on left type
        if isinstance(self._left, int):
            left_index = self._left
        elif isinstance(self._left, str):
            left_index_found = text.find(self._left)
            if left_index_found == -1:
                return ""
            left_index = (
                left_index_found
                if self._include_left
                else left_index_found + len(self._left)
            )
        else:
            left_index = 0  # default if neither int nor str

        # Determine end index based on right type
        if isinstance(self._right, int):
            right_index = self._right
        elif isinstance(self._right, str):
            right_index_found = text.rfind(self._right)
            if right_index_found == -1:
                return ""
            right_index = (
                right_index_found + len(self._right)
                if self._include_right
                else right_index_found
            )
        else:
            right_index = len(text)  # default if neither int nor str

        result = text[left_index:right_index]
        if self._strip:
            result = result.strip()
        return result



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import time

import dask_cudf

from nemo_curator import BucketsToEdges
from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import get_client, get_num_workers
from nemo_curator.utils.script_utils import ArgumentHelper


def attach_args():
    description = """
    Takes the buckets generated from minhashes and converts
    them into an edge list for the connected components algorithm. This is done by
    assuming all documents in the same bucket are similar.
    """
    parser = argparse.ArgumentParser(
        description,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.parse_gpu_dedup_args()
    parser.add_argument(
        "--input-bucket-dir",
        type=str,
        help="The directory containing anchor_docs_with_bk files.",
    )
    parser.add_argument(
        "--input-bucket-field",
        type=str,
        default="_bucket_id",
        help="Name of the column containing the bucket ID.",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        help="Output directory to write results.",
    )
    return parser


def main(args):
    logger = create_logger(
        rank=0,
        log_file=os.path.join(args.log_dir, "rank_000.log"),
        name="buckets_to_cc_log",
    )

    input_bucket_path = args.input_bucket_dir
    OUTPUT_PATH = args.output_dir

    client = get_client(**ArgumentHelper.parse_client_args(args))
    logger.info(f"Client Created {client}")
    logger.info(f"Num Workers = {get_num_workers(client)}")
    logger.info(
        "Running buckets -> EdgeList for CC",
    )

    buckets_to_edges = BucketsToEdges(
        cache_dir=OUTPUT_PATH,
        id_fields=["dataset_id", "doc_id"],
        str_id_name=args.input_json_id_field,
        bucket_field=args.input_bucket_field,
        logger=logger,
    )
    st = time.time()
    buckets_df = DocumentDataset(
        dask_cudf.read_parquet(input_bucket_path, split_row_groups=False)
    )
    _ = buckets_to_edges(buckets_df)
    et = time.time()
    logger.info(f"Bucket to Edges conversion took = {et-st} s")


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import time

import dask

from nemo_curator import FuzzyDuplicates, FuzzyDuplicatesConfig
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, write_to_disk
from nemo_curator.utils.script_utils import ArgumentHelper


def pre_imports():
    import cudf  # noqa: F401


def main(args):

    dataset_dir = "/path/to/dataset"
    log_dir = "./"
    cache_dir = "./fuzzy_cache"  # must be cleared between runs
    output_dir = "./output"
    dataset_id_field = "id"
    dataset_text_field = "text"

    filetype = "parquet"

    # Fuzzy dup calculation only supports the cuDF/GPU backend
    backend = "cudf"
    assert args.device == "gpu"

    with dask.config.set({"dataframe.backend": backend}):
        client = get_client(**ArgumentHelper.parse_client_args(args))
        client.run(pre_imports)

        t0 = time.time()
        if filetype == "parquet":
            from dask import dataframe as dd

            input_dataset = DocumentDataset(
                dd.read_parquet(
                    dataset_dir,
                    columns=[dataset_id_field, dataset_text_field],
                    blocksize="256MiB",
                    aggregate_files=True,
                )
            )
        elif filetype == "jsonl":
            input_dataset = DocumentDataset.read_json(
                dataset_dir,
                backend=backend,
            )

        fuzzy_dedup_config = FuzzyDuplicatesConfig(
            cache_dir=cache_dir,
            id_field=dataset_id_field,
            text_field=dataset_text_field,
            # Decides whether output of the module is a deduplicated dataset or the IDs of the duplicates
            perform_removal=False,
            seed=42,
            char_ngrams=24,
            num_buckets=20,
            hashes_per_bucket=13,
            use_64_bit_hash=False,
            buckets_per_shuffle=5,  # set to a smaller value if encountering OOMs during LSH
            false_positive_check=False,
        )
        fuzzy_dup = FuzzyDuplicates(logger=log_dir, config=fuzzy_dedup_config)

        # When perform_removal=False, it will only call .identify_duplicates() and return the list of duplicate IDs.
        # When perform_removal=True, then exact_dup outputs the dataset with the duplicates removed.
        # It will behave by calling .identify_duplicates() and .remove() in sequence.
        duplicates = fuzzy_dup(
            dataset=input_dataset
        )  # or fuzzy_dup.identify_duplicates(input_dataset)

        if duplicates is None:
            print("No duplicates found")
            print(f"Time taken:{time.time() - t0}s")
            return

        result = fuzzy_dup.remove(input_dataset, duplicates)
        write_to_disk(result, output_dir, output_type=filetype)
        print(f"Time taken:{time.time() - t0}s")


def attach_args(
    parser=argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    ),
):
    return ArgumentHelper(parser).add_distributed_args()


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import Iterable

import nvidia.dali.fn as fn
import nvidia.dali.types as types
import timm
import torch
from nvidia.dali import pipeline_def
from nvidia.dali.plugin.pytorch import feed_ndarray

from nemo_curator.image.embedders.base import ImageEmbedder
from nemo_curator.utils.image.transforms import convert_transforms_to_dali


class TimmImageEmbedder(ImageEmbedder):
    """
    PyTorch Image Models (timm) is a library containing SOTA computer vision
    models. Many of these models are useful in generating image embeddings
    for modules in NeMo Curator. This module can also automatically convert
    the image transformations from PyTorch transformations to DALI transformations
    in the supported models.
    """

    def __init__(
        self,
        model_name: str,
        pretrained: bool = False,
        batch_size: int = 1,
        num_threads_per_worker: int = 4,
        image_embedding_column: str = "image_embedding",
        normalize_embeddings: bool = True,
        classifiers: Iterable = [],
        autocast: bool = True,
        use_index_files: bool = False,
    ) -> None:
        """
        Constructs the embedder.

        Args:
            model_name (str): The timm model to use. A list of available models
                can be found by running timm.list_models()
            pretrained (bool): If True, loads the pretrained weights of the model.
            batch_size (int): The number of images to run inference on in a single batch.
                If the batch_size is larger than the number of elements in a shard, only
                the number of elements in a shard will be used.
            num_threads_per_worker (int): The number of threads per worker (GPU) to use
                for loading images with DALI.
            image_embedding_column (str): The output column where the embeddings will be
                stored in the dataset.
            normalize_embeddings (bool): Whether to normalize the embeddings output by the
                model. Defaults to True.
            classifiers (Iterable): A collection of classifiers to immediately apply on top
                of the image embeddings.
            autocast (bool): If True, runs the timm model using torch.autocast().
            use_index_files (bool): If True, tries to find and use index files generated
                by DALI at the same path as the tar file shards. The index files must be
                generated by DALI's wds2idx tool. See https://docs.nvidia.com/deeplearning/dali/user-guide/docs/examples/general/data_loading/dataloading_webdataset.html#Creating-an-index
                for more information. Each index file must be of the form "shard_id.idx"
                where shard_id is the same integer as the corresponding tar file for the
                data. The index files must be in the same folder as the tar files.
        """
        super().__init__(
            model_name=model_name,
            image_embedding_column=image_embedding_column,
            classifiers=classifiers,
        )
        self.pretrained = pretrained
        self.batch_size = batch_size
        self.num_threads_per_worker = num_threads_per_worker
        self.normalize_embeddings = normalize_embeddings
        self.autocast = autocast
        self.use_index_files = use_index_files

        # Load the model to get the transforms
        model = timm.create_model(self.model_name, pretrained=self.pretrained)
        torch_transforms = timm.data.create_transform(
            **timm.data.resolve_data_config(model.pretrained_cfg)
        )
        self.dali_transforms = convert_transforms_to_dali(torch_transforms)

    def load_dataset_shard(self, tar_path: str):
        """
        Loads a WebDataset tar shard using DALI.

        Args:
            tar_path (str): The path of the tar shard to load.

        Returns:
            Iterable: An iterator over the dataset. Each tar file
            must have 3 files per record: a .jpg file, a .txt file,
            and a .json file. The .jpg file must contain the image, the
            .txt file must contain the associated caption, and the
            .json must contain the metadata for the record (including
            its ID). Images will be loaded using DALI.
        """

        # Create the DALI pipeline
        @pipeline_def(
            batch_size=self.batch_size,
            num_threads=self.num_threads_per_worker,
            device_id=0,
        )
        def webdataset_pipeline(_tar_path: str):
            if self.use_index_files:
                index_paths = [f"{_tar_path.rsplit('.', 1)[0]}.idx"]
            else:
                index_paths = []

            img_raw, text, json = fn.readers.webdataset(
                paths=_tar_path,
                index_paths=index_paths,
                ext=["jpg", "txt", "json"],
                missing_component_behavior="error",
            )
            img = fn.decoders.image(img_raw, device="mixed", output_type=types.RGB)

            for transform in self.dali_transforms:
                img = transform(img)

            return img, text, json

        pipeline = webdataset_pipeline(tar_path)
        pipeline.build()

        total_samples = pipeline.epoch_size()
        total_samples = total_samples[list(total_samples.keys())[0]]

        samples_completed = 0
        while samples_completed < total_samples:
            image, text, meta = pipeline.run()
            image = image.as_tensor()

            image_torch = torch.empty(image.shape(), dtype=torch.float32, device="cuda")
            feed_ndarray(image, image_torch)  # COPY !!!
            image = image_torch

            captions = [text.at(i).tostring().decode("utf-8") for i in range(len(text))]
            metadata = [
                json.loads(meta.at(i).tostring().decode("utf-8"))
                for i in range(len(meta))
            ]

            remaining_samples = total_samples - samples_completed
            if image.shape[0] >= remaining_samples:
                image = image[:remaining_samples]
                captions = captions[:remaining_samples]
                metadata = metadata[:remaining_samples]

            samples_completed += min(image.shape[0], remaining_samples)

            yield image, metadata

    def load_embedding_model(self, device="cuda"):
        """
        Loads the model used to generate image embeddings.

        Args:
            device (str): A PyTorch device identifier that specifies what GPU
                to load the model on.

        Returns:
            Callable: A timm model loaded on the specified device.
                The model's forward call may be augmented with torch.autocast()
                or embedding normalization if specified in the constructor.
        """
        model = timm.create_model(self.model_name, pretrained=self.pretrained).eval()
        model = model.to(device)
        model = self._configure_forward(model)

        return model

    def _configure_forward(self, model):
        original_forward = model.forward

        def custom_forward(*args, **kwargs):
            if self.autocast:
                with torch.autocast(device_type="cuda"):
                    image_features = original_forward(*args, **kwargs)
            else:
                image_features = original_forward(*args, **kwargs)

            if self.normalize_embeddings:
                image_features = torch.nn.functional.normalize(image_features, dim=-1)

            # Inference can be done in lower precision, but cuDF can only handle fp32
            return image_features.to(torch.float32)

        model.forward = custom_forward
        return model



import ast
import json

from Endpoints import *
from prompts import *


class Generator:
    def __init__(self):
        self.llm = LLaMa_405B()

    def extract_points_of_interest(self, persona, file_name, passage):
        prompt = extract_user_interest_prompt.format(
            persona=persona, file_name=file_name, passage=passage
        )

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "list_of_interest": {
                        "description": "[<fill with 1-5 word desription>]",
                        "type": "array",
                    }
                },
                "required": ["list_of_interest"],
            }
        }

        raw_answer = json.loads(self.llm.invoke(prompt, schema))
        return raw_answer

    def extract_compatible_question_type(self, interest, types, file_name, passage):
        prompt = extract_compatible_question_type_prompt.format(
            interest="\n".join(interest),
            types="\n".join(types),
            file_name=file_name,
            passage=passage,
        )

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "reasoning": {
                        "description": "show your reasoning",
                        "type": "string",
                    },
                    "list_of_extractable_types_of_questions": {
                        "description": "list_of_extractable_types_of_questions",
                        "type": "array",
                    },
                },
                "required": ["reasoning", "list_of_extractable_types_of_questions"],
            }
        }
        answer = json.loads(self.llm.invoke(prompt, schema))
        return answer

    def generate_questions(self, file_name, passage, interest, types):
        prompt = extract_questions_prompt.format(
            file_name=file_name, passage=passage, interest=interest, types=types
        )

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "generated_questions": {
                        "description": "[questions]",
                        "type": "array",
                    }
                },
                "required": ["generated_questions"],
            }
        }
        try:
            answer = json.loads(self.llm.invoke(prompt, schema))["generated_questions"]
            return answer
        except:
            return []

    def conversational_re_write(self, question, file_name, passage):
        prompt = conversational_re_write_prompt.format(
            question=question, file_name=file_name, passage=passage
        )
        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "re_written_question": {"description": "<fill>", "type": "string"}
                },
                "required": ["re_written_question"],
            }
        }

        answer = json.loads(self.llm.invoke(prompt, schema))
        return answer

    def writing_style(self, persona):
        prompt = extract_writing_style.format(persona=persona)
        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "writing_style": {
                        "description": "<the writing style described in great detail in a paragraph>",
                        "type": "string",
                    }
                },
                "required": ["writing_style"],
            }
        }
        answer = self.llm.invoke(prompt, schema)

        return answer

    def persona_rewrite(self, persona, question):
        prompt = persona_rewrite_prompt.format(persona=persona, question=question)

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "new_question": {
                        "description": "<the writing style described in great detail in a paragraph>",
                        "type": "string",
                    }
                },
                "required": ["new_question"],
            }
        }
        try:
            answer = self.llm.invoke(prompt, schema)
            return answer
        except:
            return {"reasoning": "error", "new_question": question}


class Relevance_Filter:
    def __init__(self):
        self.llm = LLaMa_405B()

    def execute(self, question, file_name, passage):
        prompt = filter_relevance_prompt.format(
            question=question, file_name=file_name, passage=passage
        )

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "Reasoning": {
                        "description": "1-10 words of reasoning",
                        "type": "string",
                    },
                    "Your_Decision": {
                        "description": "fill with judgement option",
                        "type": "string",
                    },
                },
                "required": ["Reasoning", "Your_Decision"],
            }
        }

        answer = json.loads(self.llm.invoke(prompt, schema))

        return answer


class Intelligent_Question_Filter:
    def __init__(self):
        self.llm = LLaMa_405B()

    def execute(self, question, file_name, passage):
        prompt = intelligent_question_filter_prompt.format(
            question=question, file_name=file_name, passage=passage
        )

        schema = {
            "guided_json": {
                "type": "object",
                "properties": {
                    "Type_of_question": {
                        "description": "<Fill with Type_A or Type_B or Type_C>",
                        "type": "string",
                    }
                },
                "required": ["Type_of_question"],
            }
        }

        answer = json.loads(self.llm.invoke(prompt, schema))
        return answer



import argparse

from dask_kubernetes.operator.kubecluster import KubeCluster, make_cluster_spec


def create_cluster(
    name: str,
    n_workers: int,
    n_gpus_per_worker: int,
    n_cpus_per_worker: int,
    image: str,
    image_pull_secret: str,
    pvcs: dict[str, str],
):
    dask_worker_command = "dask-worker"
    if n_gpus_per_worker and n_gpus_per_worker > 0:
        dask_worker_command = "dask-cuda-worker"

    custom_cluster_spec = make_cluster_spec(
        name=name,
        worker_command=dask_worker_command,
        n_workers=n_workers,
        image=image,
    )
    scheduler_spec = custom_cluster_spec["spec"]["scheduler"]["spec"]
    worker_spec = custom_cluster_spec["spec"]["worker"]["spec"]
    if image_pull_secret:
        scheduler_spec["imagePullSecrets"] = [{"name": image_pull_secret}]
        worker_spec["imagePullSecrets"] = [{"name": image_pull_secret}]

    obj_vols = []
    obj_vol_mounts = []
    for pvc_name, mount_path in pvcs.items():
        obj_vols.append(
            {
                "name": pvc_name,
                "persistentVolumeClaim": {
                    "claimName": pvc_name,
                },
            }
        )
        obj_vol_mounts.append(
            {
                "name": pvc_name,
                "mountPath": mount_path,
            }
        )

    scheduler_spec["volumes"] = obj_vols
    for ctr in scheduler_spec["containers"]:
        ctr["volumeMounts"] = obj_vol_mounts

    worker_spec["volumes"] = obj_vols
    for ctr in worker_spec["containers"]:
        ctr["volumeMounts"] = obj_vol_mounts
        # Resources are added to only the worker, since the scheduler doesn't need GPUs
        if n_gpus_per_worker or n_cpus_per_worker:
            if not ctr["resources"]:
                ctr["resources"] = {"limits": {}}
            if n_gpus_per_worker:
                ctr["resources"]["limits"]["nvidia.com/gpu"] = str(n_gpus_per_worker)
            if n_cpus_per_worker:
                ctr["resources"]["limits"]["cpu"] = str(n_cpus_per_worker)

    cluster = KubeCluster(
        custom_cluster_spec=custom_cluster_spec, shutdown_on_close=False
    )
    print(f"{cluster = }")


if __name__ == "__main__":

    def parse_pvcs(specs: str) -> dict[str, str]:
        name_to_path = {}
        for pvc in specs.split(","):
            # Can be empty
            if not pvc:
                continue
            name, _, path = pvc.partition(":")
            name_to_path[name] = path
        return name_to_path

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--name",
        type=str,
        default="rapids-dask",
        help="The name of the DaskCluster which you would be able to inspect via `kubectl describe daskcluster <name>`.",
    )
    parser.add_argument(
        "-w", "--n_workers", type=int, default=2, help="Number of workers"
    )
    parser.add_argument(
        "-g",
        "--n_gpus_per_worker",
        type=int,
        default=None,
        help="Number of GPUs per worker. If not specified, the Dask Cluster defaults to a CPU cluster.",
    )
    parser.add_argument(
        "-c",
        "--n_cpus_per_worker",
        type=int,
        default=None,
        help="Number of CPUs per worker. Provide this flag if you want to limit your CPU resources and K8s will throttle the workers to make sure this limit is satisfied.",
    )
    parser.add_argument(
        "-i",
        "--image",
        type=str,
        default="nvcr.io/nvidia/nemo:24.03.framework",
        help="The image used for the Dask Cluster scheduler and workers.",
    )
    parser.add_argument(
        "-s",
        "--image_pull_secret",
        type=str,
        default=None,
        help="If --image is from a private registry, specify the appropriate pull secret you created to allow these to be pulled.",
    )
    parser.add_argument(
        "-p",
        "--pvcs",
        type=parse_pvcs,
        default="",
        help="Comma sep PVC specificiation of $pvc_name_1:$mount_path_1,$pvc_name_2:$mount_path_2. Example: foo:/foo,bar:/bar mounts pvcs named foo and bar to /foo and /bar respectively.",
    )

    args = parser.parse_args()

    create_cluster(
        **vars(args),
    )



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


MAJOR = 0
MINOR = 8
PATCH = 0
PRE_RELEASE = ""
DEV = ""

# Use the following formatting: (major, minor, patch, pre-release)
VERSION = (MAJOR, MINOR, PATCH, PRE_RELEASE, DEV)

__shortversion__ = ".".join(map(str, VERSION[:3]))
__version__ = __shortversion__

if VERSION[3] != "":
    __version__ = __version__ + VERSION[3]

if VERSION[4] != "":
    __version__ = __version__ + "." + ".".join(VERSION[4:])

__package_name__ = "nemo_curator"
__contact_names__ = "NVIDIA"
__contact_emails__ = "nemo-toolkit@nvidia.com"
__homepage__ = (
    "https://docs.nvidia.com/nemo-framework/user-guide/latest/datacuration/index.html"
)
__repository_url__ = "https://github.com/NVIDIA/NeMo-Curator"
__download_url__ = "https://github.com/NVIDIA/NeMo-Curator/releases"
__description__ = (
    "NeMo Curator - Scalable Data Preprocessing Tool for Training Large Language Models"
)
__license__ = "Apache2"
__keywords__ = "deep learning, machine learning, gpu, NLP, NeMo, nvidia, pytorch, torch, language, preprocessing, LLM, large language model"



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .arxiv import ArxivDownloader, ArxivExtractor, ArxivIterator, download_arxiv
from .commoncrawl import (
    CommonCrawlWARCDownloader,
    CommonCrawlWARCDownloaderExtractOnly,
    CommonCrawlWARCExtractor,
    CommonCrawlWARCIterator,
    JusTextExtractor,
    ResiliparseExtractor,
    TrafilaturaExtractor,
    download_common_crawl,
)
from .doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
    batch_download,
    download_and_extract,
    import_downloader,
    import_extractor,
    import_iterator,
)
from .wikipedia import (
    WikipediaDownloader,
    WikipediaExtractor,
    WikipediaIterator,
    download_wikipedia,
)

__all__ = [
    "DocumentDownloader",
    "DocumentIterator",
    "DocumentExtractor",
    "download_and_extract",
    "import_downloader",
    "import_extractor",
    "import_iterator",
    "download_common_crawl",
    "CommonCrawlWARCDownloader",
    "CommonCrawlWARCExtractor",
    "CommonCrawlWARCIterator",
    "CommonCrawlWARCDownloaderExtractOnly",
    "JusTextExtractor",
    "ResiliparseExtractor",
    "TrafilaturaExtractor",
    "download_wikipedia",
    "WikipediaDownloader",
    "WikipediaIterator",
    "WikipediaExtractor",
    "batch_download",
    "download_arxiv",
    "ArxivDownloader",
    "ArxivIterator",
    "ArxivExtractor",
]



import logging
import os

from dask.distributed import Client, LocalCluster
from helper import process_data

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)

DATA_BASE = os.environ.get("DATA_BASE")
INPUT_BASE = os.path.join(DATA_BASE, "raw/data/zyda_no_starcoder")
OUTPUT_BASE = os.path.join(DATA_BASE, "processed/zyda-parquet")
CPU_WORKERS = os.environ.get("CPU_WORKERS")


if __name__ == "__main__":
    logging.info("Starting Dask cluster")
    cluster = LocalCluster(n_workers=CPU_WORKERS, processes=True, memory_limit="48GB")
    client = Client(cluster)
    logging.info(client)

    components = [
        "zyda_arxiv",
        "zyda_peS2o",
        "zyda_pile-uncopyrighted",
        "zyda_slimpajama",
        "zyda_c4-en",
        "zyda_refinedweb",
    ]

    for component in components:
        input_path = os.path.join(INPUT_BASE, component)
        if not os.path.exists(input_path):
            continue
        output_path = os.path.join(OUTPUT_BASE, component)
        logging.info(f"Processing {component}")
        process_data(
            input_folder=input_path, output_folder=output_path, prefix=component
        )
        logging.info("Done!")

    client.cluster.close()
    client.shutdown()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import cudf
import dask_cuda
import numpy as np
from dask import config
from packaging.version import Version

from nemo_curator._compat import query_planning_enabled
from nemo_curator.utils.fuzzy_dedup_utils.output_map_utils import build_partition

dask_cuda_version = Version(dask_cuda.__version__)
USE_EXCOMMS = (
    dask_cuda_version >= Version("23.10") and dask_cuda_version < Version("24.06")
) or dask_cuda_version >= Version("24.08")


def write_partitioned_file(df, output_path, partition_on, batch_id):
    if len(df) == 0:
        return cudf.Series([True])

    cudf.io.parquet.write_to_dataset(
        df,
        output_path,
        partition_cols=[partition_on],
        filename=f"batch_{batch_id}.parquet",
    )
    return cudf.Series([True])


def rearange_by_column_direct(
    df,
    col,
    npartitions,
    ignore_index,
    excomms_default=USE_EXCOMMS,
):
    # Execute a "direct" shuffle operation without staging
    if config.get("explicit-comms", excomms_default):
        from dask_cuda.explicit_comms.dataframe.shuffle import (
            shuffle as explicit_comms_shuffle,
        )

        # Use explicit comms unless the user has
        # disabled it with the dask config system,
        # or we are using an older version of dask-cuda
        return explicit_comms_shuffle(
            df,
            [col],
            npartitions=npartitions,
            ignore_index=ignore_index,
        )

    elif query_planning_enabled():
        try:
            from dask.dataframe import dask_expr
        except ImportError:
            # TODO: Remove when pinned to dask>2024.12.1
            import dask_expr

        # Use the internal dask-expr API
        return dask_expr.new_collection(
            dask_expr._shuffle.RearrangeByColumn(
                frame=df.expr,
                partitioning_index=col,
                npartitions_out=npartitions,
                ignore_index=ignore_index,
                method="tasks",
                # Prevent staged shuffling by setting max_branch
                # to the number of input partitions + 1
                options={"max_branch": npartitions + 1},
            )
        )

    else:
        from dask.dataframe.shuffle import rearrange_by_column

        return rearrange_by_column(
            df,
            col=col,
            shuffle_method="tasks",
            # Prevent staged shuffling by setting max_branch
            # to the number of input partitions + 1
            max_branch=npartitions + 1,
            npartitions=npartitions,
            ignore_index=ignore_index,
        )


def get_shuffle_part_ids_df(
    agg_df,
    partition_on,
    output_col,
    size_col,
    num_workers=0,
):
    sizes = agg_df[size_col].values
    max_text_bytes_per_part = int(np.iinfo(np.int32).max * 3)

    # Adjust max_text_bytes_per_part if the number of output
    # partitions is small compared to the number of workers.
    # Sometimes we just have very few output partitions to
    # deal with, and just need a larger batch
    npartitions_min = max(1, int(num_workers * 0.8))
    while True:
        output_ar = build_partition(sizes.get(), max_text_bytes_per_part)
        if output_ar.max() > npartitions_min or max_text_bytes_per_part < 2**24:
            break
        max_text_bytes_per_part = int(max_text_bytes_per_part // 2.0)

    df = cudf.DataFrame()
    df[partition_on] = agg_df[partition_on]
    df[output_col] = output_ar
    return df



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Set, Tuple

import requests

from nemo_curator.download.doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
)


class TinyStoriesDownloader(DocumentDownloader):
    def __init__(self, download_dir: str):
        super().__init__()

        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)

        self._download_dir = download_dir
        print("Download directory: ", self._download_dir)

    def download(self, url: str) -> str:
        filename = os.path.basename(url)
        output_file = os.path.join(self._download_dir, filename)

        if os.path.exists(output_file):
            print(f"File '{output_file}' already exists, skipping download.")
            return output_file

        print(f"Downloading TinyStories dataset from '{url}'...")
        response = requests.get(url)

        with open(output_file, "wb") as file:
            file.write(response.content)

        return output_file


class TinyStoriesIterator(DocumentIterator):
    # The token that separates stories in the TinyStories dataset.
    SEPARATOR_TOKEN = "<|endoftext|>"

    def __init__(self):
        super().__init__()
        self._counter = -1

    def iterate(self, file_path):
        self._counter = -1
        file_name = os.path.basename(file_path)

        with open(file_path, "r") as file:
            example = []

            def split_meta(example):
                if example:
                    self._counter += 1
                    content = " ".join(example)
                    meta = {
                        "file_name": file_name,
                        "id": f"{file_name}-{self._counter}",
                    }

                    return meta, content

            for line in file:
                if line.strip() == TinyStoriesIterator.SEPARATOR_TOKEN:
                    if example:
                        yield split_meta(example)
                        example = []
                else:
                    example.append(line.strip())

            if example:
                yield split_meta(example)


class TinyStoriesExtractor(DocumentExtractor):
    def extract(self, content: str) -> Tuple[Set, str]:
        # No metadata for the text, just the content.
        return {}, content



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

import dask.dataframe
import pandas as pd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers.pii_modifier import PiiModifier
from nemo_curator.modules.modify import Modify
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper


def console_script():
    parser = argparse.ArgumentParser()
    args = ArgumentHelper(parser).add_distributed_args().parse_args()
    _ = get_client(**ArgumentHelper.parse_client_args(args))

    dataframe = pd.DataFrame(
        {"text": ["Sarah and Ryan went out to play", "Jensen is the CEO of NVIDIA"]}
    )
    dataset = DocumentDataset.from_pandas(dataframe, npartitions=1)

    modifier = PiiModifier(
        log_dir="./logs",
        batch_size=2000,
        language="en",
        supported_entities=["PERSON", "EMAIL_ADDRESS"],
        anonymize_action="replace",
    )

    modify = Modify(modifier)
    modified_dataset = modify(dataset)
    modified_dataset.to_json("output.jsonl")


if __name__ == "__main__":
    console_script()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nemo_curator.datasets import DocumentDataset
from nemo_curator.modifiers import DocumentModifier
from nemo_curator.modules.base import BaseModule
from nemo_curator.utils.module_utils import is_batched


class Modify(BaseModule):
    def __init__(self, modifier: DocumentModifier, text_field="text"):
        super().__init__(input_backend=modifier.backend)
        self.modifier = modifier
        self.text_field = text_field

    def call(self, dataset: DocumentDataset) -> DocumentDataset:
        if is_batched(self.modifier.modify_document):
            dataset.df[self.text_field] = dataset.df[self.text_field].map_partitions(
                self.modifier.modify_document, meta=(None, str)
            )
        else:
            dataset.df[self.text_field] = dataset.df[self.text_field].apply(
                self.modifier.modify_document, meta=(None, str)
            )

        return dataset



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import pickle

import nemo_curator
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.distributed_utils import get_client, read_data, write_to_disk
from nemo_curator.utils.file_utils import (
    expand_outdir_and_mkdir,
    get_all_files_paths_under,
    get_batched_files,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    output_tdd_dir = expand_outdir_and_mkdir(args.output_task_deduped_dir)
    output_rm_doc_dir = None
    if args.output_removed_doc_dir is not None:
        output_rm_doc_dir = expand_outdir_and_mkdir(args.output_removed_doc_dir)

    # Each rank read in the task data
    print(f"Reading in matched n-grams from {args.input_matched_ngrams}")
    with open(args.input_matched_ngrams, "rb") as fp:
        matched_ngram_data = pickle.load(fp)

    # Unpack the results from find_matched_ngrams
    matched_ngrams = matched_ngram_data["matched-ngrams"]
    ngrams_freq = matched_ngram_data["ngrams-freq"]
    max_ngram_size = matched_ngram_data["max-ngram-size"]

    decontaminator = nemo_curator.TaskDecontamination(
        [],
        text_field=args.input_text_field,
        max_ngram_size=max_ngram_size,
        max_matches=args.match_threshold,
        max_splits=args.max_document_splits,
        removed_dir=output_rm_doc_dir,
    )

    files = list(get_all_files_paths_under(args.input_data_dir))
    for files in get_batched_files(
        args.input_data_dir,
        output_tdd_dir,
        args.input_file_type,
        batch_size=args.batch_size,
    ):
        dataset = DocumentDataset(
            read_data(
                files,
                file_type=args.input_file_type,
                backend="pandas",
                add_filename=True,
            )
        )
        decontaminated_dataset = decontaminator.remove_matching_ngrams(
            matched_ngrams, ngrams_freq, dataset
        )
        write_to_disk(
            decontaminated_dataset.df,
            output_tdd_dir,
            write_to_filename=True,
            output_type=args.output_file_type,
        )
        print(f"Finished decontaminating {len(files)} files")

    print("Finished decontaminating all files")


def attach_args(
    parser=argparse.ArgumentParser(
        """
 Using the matching n-grams find by
 nemo_curator/scripts/find_matching_ngrams.py
 (provided by the argument --input-matched-ngrams),
 passes over all documents and removes matching n-grams from the corpus by
 splitting documents containing the match. If a document is split more than
 --max-splits times, it is removed from the corpus.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_batch_size()
    argumentHelper.add_arg_input_data_dir()
    argumentHelper.add_arg_input_file_type()
    argumentHelper.add_arg_input_text_field()
    argumentHelper.add_arg_output_file_type()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--input-matched-ngrams",
        type=str,
        default=None,
        required=True,
        help="Input dictionary (.pkl file) that contains matched "
        "n-gram data from the find_matching_ngrams code.",
    )
    parser.add_argument(
        "--match-threshold",
        type=int,
        default=10,
        help="A threshold that determines if a matched n-gram will be "
        "considered for removal in remove_matching_ngrams. N-grams that "
        "exceed this number of matches in the training dataset will not be "
        "considered during the removal stage.",
    )
    parser.add_argument(
        "--max-document-splits",
        type=int,
        default=10,
        help="A threshold used to determine if a document should be removed "
        "from the corpus if it is split more than "
        "--max-document-splits number of times.",
    )
    parser.add_argument(
        "--output-removed-doc-dir",
        type=str,
        default=None,
        help="Output directory to where removed documents will be written. "
        "Documents will be removed from the corpus if they are split more "
        "than --max-document-splits number of times, or if the user specifies "
        "that they be removed via the flag --remove-split-docs.",
    )
    parser.add_argument(
        "--output-task-deduped-dir",
        type=str,
        default=None,
        required=True,
        help="Output directory to where task-deduplicated (split) "
        "documents will be written.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



import logging
import os
import time

from nemo_curator import ConnectedComponents
from nemo_curator.utils.distributed_utils import get_client, get_num_workers

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)


DATA_BASE = os.environ.get("DATA_BASE")
SCHEDULER_FILE = os.environ.get("SCHEDULER_FILE")


if __name__ == "__main__":
    client = get_client(scheduler_file=SCHEDULER_FILE)
    logging.info(f"Number of dask workers: {get_num_workers(client)}")
    # Input
    buckets_to_edges_out = os.path.join(
        DATA_BASE, "fuzzy/buckets_to_edges/data/_edges.parquet"
    )

    # Output
    connected_component_base_output_path = os.path.join(DATA_BASE, "fuzzy/cc")
    connected_component_output_path = os.path.join(
        connected_component_base_output_path, "connected_components.parquet"
    )
    connected_component_cache_dir = os.path.join(
        connected_component_base_output_path, "cache"
    )

    # Relevant parameters
    input_id_field = "id"

    t0 = time.time()

    components_stage = ConnectedComponents(
        cache_dir=connected_component_cache_dir,
        jaccard_pairs_path=buckets_to_edges_out,
        id_column=input_id_field,
    )

    # Load and run connected components
    components_stage.cc_workflow(output_path=connected_component_output_path)
    logging.info(f"Time taken for Connected Components: {time.time() - t0:.2f} s")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .async_nemotron import AsyncNemotronGenerator
from .async_nemotron_cc import AsyncNemotronCCGenerator
from .error import YamlConversionError
from .mixtral import Mixtral8x7BFormatter
from .nemotron import NemotronFormatter, NemotronGenerator
from .nemotron_cc import (
    NemotronCCDiverseQAPostprocessor,
    NemotronCCGenerator,
    NemotronCCKnowledgeListPostprocessor,
)
from .no_format import NoFormat
from .prompts import (
    DEFAULT_CLOSED_QA_PROMPT_TEMPLATE,
    DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_MATH_CLASSIFICATION_PROMPT_TEMPLATE,
    DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_CLASSIFICATION_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE,
    DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE,
    DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE,
    DEFAULT_SUBTOPICS_PROMPT_TEMPLATE,
    DEFAULT_WRITING_TASK_PROMPT_TEMPLATE,
    DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE,
    DIALOGUE_COMPLEX_USER_TURN_PROMPT_TEMPLATE,
    DIALOGUE_CONCISE_USER_TURN_PROMPT_TEMPLATE,
    DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE,
    MATH_PROBLEM_BEGINNER_PROMPT_TEMPLATE,
    MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE,
    PYTHON_PROBLEM_ADVANCED_PROMPT_TEMPLATE,
    PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE,
    PYTHON_PROBLEM_INTERMEDIATE_PROMPT_TEMPLATE,
)

__all__ = [
    "NemotronGenerator",
    "AsyncNemotronGenerator",
    "NemotronFormatter",
    "NemotronCCGenerator",
    "AsyncNemotronCCGenerator",
    "NemotronCCDiverseQAPostprocessor",
    "NemotronCCKnowledgeListPostprocessor",
    "Mixtral8x7BFormatter",
    "NoFormat",
    "DEFAULT_MACRO_TOPICS_PROMPT_TEMPLATE",
    "DEFAULT_SUBTOPICS_PROMPT_TEMPLATE",
    "DEFAULT_OPEN_QA_FROM_TOPICS_PROMPT_TEMPLATE",
    "DEFAULT_YAML_CONVERSION_PROMPT_TEMPLATE",
    "DEFAULT_REVISE_OPEN_QA_PROMPT_TEMPLATE",
    "DEFAULT_WRITING_TASK_PROMPT_TEMPLATE",
    "DEFAULT_REVISE_WRITING_TASK_PROMPT_TEMPLATE",
    "DEFAULT_CLOSED_QA_PROMPT_TEMPLATE",
    "DEFAULT_MATH_MACRO_TOPICS_PROMPT_TEMPLATE",
    "DEFAULT_MATH_SUBTOPICS_PROMPT_TEMPLATE",
    "DEFAULT_MATH_CLASSIFICATION_PROMPT_TEMPLATE",
    "MATH_PROBLEM_GENERAL_PROMPT_TEMPLATE",
    "MATH_PROBLEM_BEGINNER_PROMPT_TEMPLATE",
    "DEFAULT_PYTHON_MACRO_TOPICS_PROMPT_TEMPLATE",
    "DEFAULT_PYTHON_SUBTOPICS_PROMPT_TEMPLATE",
    "DEFAULT_PYTHON_CLASSIFICATION_PROMPT_TEMPLATE",
    "PYTHON_PROBLEM_BEGINNER_PROMPT_TEMPLATE",
    "PYTHON_PROBLEM_INTERMEDIATE_PROMPT_TEMPLATE",
    "PYTHON_PROBLEM_ADVANCED_PROMPT_TEMPLATE",
    "DIALOGUE_NORMAL_USER_TURN_PROMPT_TEMPLATE",
    "DIALOGUE_COMPLEX_USER_TURN_PROMPT_TEMPLATE",
    "DIALOGUE_CONCISE_USER_TURN_PROMPT_TEMPLATE",
    "YamlConversionError",
]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os

import dask
import numpy as np
import pandas as pd
import pytest
from dask import dataframe as dd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.filters import AnswerabilityFilter, DocumentFilter, EasinessFilter
from nemo_curator.modules import Filter, Score, ScoreFilter, Sequential

config_module = importlib.import_module(
    "tutorials.nemo-retriever-synthetic-data-generation.config.config"
)


@pytest.fixture
def get_original_data():
    docs = [
        {
            "_id": "930220d64a44c223df83e0caf09013fffdf4c19c1f501f035862984979928b29",
            "text": "The Eiffel Tower is an iconic landmark of Paris, France. It was designed by the engineer Gustave Eiffel and built for the 1889 Exposition Universelle (World's Fair) to celebrate the 100th anniversary of the French Revolution.",
            "title": "Eiffel Tower - A French Icon",
        },
        {
            "_id": "5cdca9fa81b6c4d8a1a1159610c98b2bffae498dad36c90639413bf22e5a4154",
            "text": "The Great Wall of China is a series of fortifications made of stone, brick, tamped earth, wood, and other materials, generally built along an east-to-west line across the historical northern borders of China to protect the Chinese states and empires against raids and invasions from various nomadic groups.",
            "title": "The Great Wall of China - Ancient Protection",
        },
    ]
    return DocumentDataset.from_pandas(pd.DataFrame(docs))


@pytest.fixture
def get_generated_data():
    docs = [
        {
            "_id": "930220d64a44c223df83e0caf09013fffdf4c19c1f501f035862984979928b29",
            "text": "The Eiffel Tower is an iconic landmark of Paris, France. It was designed by the engineer Gustave Eiffel and built for the 1889 Exposition Universelle (World's Fair) to celebrate the 100th anniversary of the French Revolution.",
            "title": "Eiffel Tower - A French Icon",
            "question-id": "d9be8cb0693a354b2ba8ddd1e86c9df57db97f33e03cb33c972f8efed4084f8b",
            "question": "What is the significance of the Eiffel Tower in relation to the French Revolution?",
            "answer": "The Eiffel Tower was built to celebrate the 100th anniversary of the French Revolution.",
        },
        {
            "_id": "5cdca9fa81b6c4d8a1a1159610c98b2bffae498dad36c90639413bf22e5a4154",
            "text": "The Great Wall of China is a series of fortifications made of stone, brick, tamped earth, wood, and other materials, generally built along an east-to-west line across the historical northern borders of China to protect the Chinese states and empires against raids and invasions from various nomadic groups.",
            "title": "The Great Wall of China - Ancient Protection",
            "question-id": "e1f6e179a883a7f108d566a582a159322b2eb2b8e0d51fd78cafc72373e4be2b",
            "question": "What is the purpose of the Great Wall of China?",
            "answer": "The purpose of the Great Wall of China is to protect the Chinese states and empires against raids and invasions from various nomadic groups.",
        },
        {
            "_id": "35f822b0b38de133b815139affac94d57d0c7d35de6e11d0e52a69d416c1d248",
            "text": "Machu Picchu is a 15th-century Inca citadel situated on a mountain ridge above the Sacred Valley in Peru. It is the most famous icon of Inca civilization, known for its sophisticated dry-stone walls that fuse huge blocks without the use of mortar.",
            "title": "Machu Picchu - Lost City of the Incas",
            "question-id": "1a60e50066c938784db3d49c41e470c197b3fc30afa07957575a1d8a34a34230",
            "question": "What is Machu Picchu renowned for in terms of its architecture?",
            "answer": "Machu Picchu is renowned for its sophisticated dry-stone walls that fuse huge blocks without the use of mortar.",
        },
    ]
    return DocumentDataset.from_pandas(pd.DataFrame(docs))


@pytest.fixture
def get_config():
    cfg = config_module.RetrieverEvalSDGConfig.from_yaml(
        "./tutorials/nemo-retriever-synthetic-data-generation/config/config.yaml"
    )
    cfg.api_key = os.environ.get("NVIDIA_API_KEY")
    return cfg


class TestSDGFilterModule:
    def test_easiness_filter(self, get_generated_data, get_config):

        ef = EasinessFilter(
            get_config.base_url,
            get_config.api_key,
            get_config.easiness_filter,
            get_config.percentile,
            get_config.truncate,
            get_config.batch_size,
        )
        easiness_filter = ScoreFilter(
            ef, text_field=["text", "question"], score_field="easiness_scores"
        )

        org_df = get_generated_data.df.compute()
        filtered_dataset = easiness_filter(get_generated_data)
        filtered_df = filtered_dataset.df.compute()
        assert "easiness_scores" in filtered_df
        assert org_df.shape[0] >= filtered_df.shape[0]

    def test_answerability_filter(self, get_generated_data, get_config):

        af = AnswerabilityFilter(
            get_config.base_url,
            get_config.api_key,
            get_config.answerability_filter,
            get_config.answerability_system_prompt,
            get_config.answerability_user_prompt_template,
            get_config.num_criteria,
        )
        answerability_filter = ScoreFilter(
            af, text_field=["text", "question"], score_field="answerability_scores"
        )
        org_df = get_generated_data.df.compute()
        filtered_dataset = answerability_filter(get_generated_data)
        filtered_df = filtered_dataset.df.compute()
        assert "answerability_scores" in filtered_df
        assert org_df.shape[0] >= filtered_df.shape[0]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from docbuilder import TinyStoriesExtractor, TinyStoriesIterator


def write_jsonl(input_filename: str, output_dir: str, dump_every_n: int = 10000):
    """
    Convert a file to JSONL format and write it to the specified output directory.

    Args:
        input_filename (str): The path to the input file.
        output_dir (str): The path to the output directory.
        dump_every_n (int, optional): The number of records to dump to file at once.
    """
    if os.path.isdir(output_dir):
        if len(os.listdir(output_dir)) > 0:
            print(
                f"Output directory '{output_dir}' is not empty. Skipping conversion to JSONL."
            )
            return
    else:
        os.makedirs(output_dir)

    print(f"Converting '{input_filename}' to JSONL format (output: '{output_dir}')...")

    basename = os.path.basename(input_filename)
    iterator = TinyStoriesIterator()
    extractor = TinyStoriesExtractor()
    to_dump = []
    dump_ctr = 0

    def dump_to_file(to_dump, dump_ctr):
        """Helper function to facilitate dumping to file."""
        output_filename = f"{basename}-{dump_ctr}.jsonl"
        with open(os.path.join(output_dir, output_filename), "w") as output_file:
            output_file.writelines(to_dump)
        # Empty out the list and increment the counter.
        return [], dump_ctr + 1

    for item in iterator.iterate(input_filename):
        record_meta, content = item
        extracted = extractor.extract(content)

        if extracted is None:
            continue

        text_meta, text = extracted

        if text is None:
            continue

        line = {
            "text": text,
            **text_meta,
            **record_meta,
        }
        json_out = json.dumps(line, ensure_ascii=False)
        to_dump.append(json_out + "\n")

        # Should we dump what we have so far?
        if len(to_dump) == dump_every_n:
            to_dump, dump_ctr = dump_to_file(to_dump, dump_ctr)

    # Dump the remaining records.
    if to_dump:
        dump_to_file(to_dump, dump_ctr)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from itertools import combinations
from typing import Iterable

import dask.dataframe as dd
import numpy as np
import pytest
import yaml
from dask import config
from dask.dataframe.utils import assert_eq

from nemo_curator import LSH, FuzzyDuplicates, FuzzyDuplicatesConfig, MinHash
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.fuzzy_dedup_utils.merge_utils import extract_partitioning_index
from nemo_curator.utils.import_utils import gpu_only_import

cudf = gpu_only_import("cudf")
dask_cudf = gpu_only_import("dask_cudf")


@pytest.fixture
def fuzzy_dedup_data():
    df = cudf.DataFrame(
        {
            "id": [1, 2, 300, 4, -1],
            "text": [
                "A test string",
                "A different test string",
                "A different object",
                "The quick brown fox jumps over the lazy dog",
                "The quick black cat jumps over the lazy dog",
            ],
        }
    )
    df = dask_cudf.from_cudf(df, 2)
    return DocumentDataset(df)


@pytest.fixture
def large_fuzzy_dedup_data():
    df = cudf.DataFrame(
        {
            "id": np.arange(500),
            "text": [
                "A test string",
                "A different test string",
                "A different object",
                "The quick brown fox jumps over the lazy dog",
                "The quick black cat jumps over the lazy dog",
            ]
            * 100,
        }
    )
    df = dask_cudf.from_cudf(df, 5).reset_index(drop=True)
    return DocumentDataset(df)


@pytest.fixture
def no_duplicates_fuzzy_dedup_data():
    df = cudf.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "text": [
                "A test string",
                "Very different thing",
                "Something completely else that doesn't match",
                "The quick black cat jumps over the lazy dog",
            ],
        }
    )
    df = dask_cudf.from_cudf(df, 2)
    return DocumentDataset(df)


@pytest.fixture
def shuffle_fail_fuzzy_dedup_data():
    df = cudf.DataFrame(
        {
            "id": [1, 2, 300, 4, -1],
            "text": [
                "A test string",
                "A different test string",
                "A different object",
                "Something completely else that doesn't match",
                "The quick black cat jumps over the lazy dog",
            ],
        }
    )
    df = dask_cudf.from_cudf(df, 2)
    return DocumentDataset(df)


def minhash_overlap(minhash1: np.array, minhash2: np.array):
    assert len(minhash1) == len(minhash2)
    overlap = sum(minhash1 == minhash2)
    return overlap / len(minhash1)


def jaccard_index(str1: str, str2: str, char_ngrams):
    return (
        cudf.Series([str1])
        .str.jaccard_index(cudf.Series([str2]), width=char_ngrams)
        .values_host[0]
    )


def generate_all_pairs(item: Iterable):
    return combinations(item, 2)


@pytest.mark.gpu
class TestMinhashes:
    @pytest.mark.parametrize("use_64bit_hash", [False, True])
    @pytest.mark.parametrize("seed,char_ngrams,num_hashes", [(128, 3, 260)])
    def test_identical_minhash(
        self, fuzzy_dedup_data, use_64bit_hash, seed, char_ngrams, num_hashes
    ):
        minhasher1 = MinHash(
            seed=seed,
            num_hashes=num_hashes,
            char_ngrams=char_ngrams,
            use_64bit_hash=use_64bit_hash,
        )
        minhash_sig1 = minhasher1(fuzzy_dedup_data)
        sig_lengths = minhash_sig1.df["_minhash_signature"].compute().list.len()
        assert (sig_lengths == num_hashes).all()

        minhasher2 = MinHash(
            seed=seed,
            num_hashes=num_hashes,
            char_ngrams=char_ngrams,
            use_64bit_hash=use_64bit_hash,
        )
        minhash_sig2 = minhasher2(fuzzy_dedup_data)
        assert_eq(minhash_sig1.df, minhash_sig2.df)

    @pytest.mark.parametrize(
        "use_64bit_hash,seed,char_ngrams,num_hashes",
        [(False, 42, 5, 20), (True, 32768, 10, 18)],
    )
    def test_minhash_approximation(
        self, fuzzy_dedup_data, use_64bit_hash, seed, char_ngrams, num_hashes
    ):
        THRESHOLD = 0.15

        minhasher = MinHash(
            seed=seed,
            num_hashes=num_hashes,
            char_ngrams=char_ngrams,
            use_64bit_hash=use_64bit_hash,
        )
        minhashes = minhasher(fuzzy_dedup_data)
        minhash_signatures = (
            minhashes.df["_minhash_signature"].compute().to_pandas().values
        )
        strings = fuzzy_dedup_data.df["text"].compute().to_pandas().values
        for (sig1, str1), (sig2, str2) in generate_all_pairs(
            tuple(zip(minhash_signatures, strings))
        ):
            true_jaccard = jaccard_index(str1, str2, char_ngrams)
            minhash_approximation = minhash_overlap(np.array(sig1), np.array(sig2))
            assert abs(true_jaccard - minhash_approximation) < THRESHOLD

    def test_minhash_cache(self, fuzzy_dedup_data, tmpdir):
        minhasher = MinHash(cache_dir=tmpdir)
        result = minhasher(fuzzy_dedup_data)
        assert len(result) == len(fuzzy_dedup_data)
        assert "_minhashes.parquet" in os.listdir(tmpdir)
        assert len(os.listdir(tmpdir / "_minhashes.parquet")) != 0


@pytest.mark.gpu
class TestLSH:
    @pytest.fixture(autouse=True)
    def minhash_data(self):
        df = cudf.DataFrame(
            {
                "id": [1, 2, 3, 4, 5],
                "dataset_id": [1, 1, 2, 3, 4],
                "minhash_sig": [
                    [1, 2, 1, 2, 1, 2],
                    [1, 2, 3, 4, 5, 6],
                    [3, 2, 1, 4, 5, 6],
                    [9, 8, 7, 6, 5, 4],
                    [3, 1, 2, 4, 5, 4],
                ],
            }
        )
        df = dask_cudf.from_cudf(df, 2)
        self.dataset = DocumentDataset(df)

    @pytest.mark.parametrize("buckets_per_shuffle", [1, 2, 3])
    def test_lsh(self, tmpdir, buckets_per_shuffle):
        lsh = LSH(
            cache_dir=tmpdir,
            num_hashes=6,
            num_buckets=3,
            buckets_per_shuffle=buckets_per_shuffle,
            minhash_field="minhash_sig",
            id_fields="id",
        )
        buckets = lsh(self.dataset)
        buckets_df = buckets.df
        docs_list = buckets_df.groupby("_bucket_id").id.agg(list)
        expected_df = cudf.Series([[1, 2], [2, 3], [4, 5]], name="id")
        assert_eq(expected_df, docs_list, check_index=False)

    def test_multiple_id_cols(self, tmpdir):
        lsh = LSH(
            cache_dir=tmpdir,
            num_hashes=6,
            num_buckets=3,
            buckets_per_shuffle=1,
            id_fields=["id", "dataset_id"],
            minhash_field="minhash_sig",
        )
        buckets = lsh(self.dataset)
        buckets_df = buckets.df.compute().to_pandas()
        buckets_df["new_id"] = list(
            map(list, zip(buckets_df.dataset_id, buckets_df.id))
        )
        docs_list = buckets_df.groupby("_bucket_id").new_id.apply(list)
        expected_df = cudf.Series(
            [[(1, 1), (1, 2)], [(1, 2), (2, 3)], [(3, 4), (4, 5)]], name="new_id"
        )
        assert_eq(expected_df, docs_list, check_index=False)

    @pytest.mark.parametrize("false_positive_check", [True, False])
    def test_no_duplicates(self, tmpdir, false_positive_check):
        minhash_df = cudf.DataFrame(
            {
                "id": [1, 2, 3, 4, 5],
                "minhash_sig": [
                    [1, 2, 1, 2, 1],
                    [2, 3, 3, 4, 5],
                    [3, 4, 5, 5, 6],
                    [4, 8, 7, 6, 7],
                    [5, 10, 9, 7, 8],
                ],
            }
        )
        minhash_dataset = DocumentDataset(dask_cudf.from_cudf(minhash_df, 2))

        lsh = LSH(
            cache_dir=tmpdir,
            num_hashes=5,
            num_buckets=5,
            buckets_per_shuffle=1,
            id_fields="id",
            minhash_field="minhash_sig",
            false_positive_check=false_positive_check,
        )
        buckets = lsh(minhash_dataset)
        assert buckets is None
        assert "_buckets.parquet" not in os.listdir(tmpdir)

    @pytest.mark.parametrize("false_positive_check", [True, False])
    def test_partial_overlap(self, tmpdir, false_positive_check):
        minhash_df = cudf.DataFrame(
            {
                "id": [1, 2, 3],
                "minhash_sig": [
                    [1, 2, 1, 1, 1],
                    [2, 3, 1, 2, 2],
                    [3, 4, 2, 3, 1],
                ],
            }
        )
        minhash_dataset = DocumentDataset(dask_cudf.from_cudf(minhash_df, 2))

        lsh = LSH(
            cache_dir=tmpdir,
            num_hashes=5,
            num_buckets=5,
            buckets_per_shuffle=1,
            id_fields="id",
            minhash_field="minhash_sig",
            false_positive_check=false_positive_check,
        )
        buckets = lsh(minhash_dataset)
        assert len(buckets) == 4
        assert buckets.df["_bucket_id"].nunique().compute() == 2
        assert_eq(
            buckets.df["id"], cudf.Series([1, 2, 1, 3], name="id"), check_index=False
        )


@pytest.mark.gpu
class TestFuzzyDuplicates:
    @pytest.mark.parametrize("use_64_bit_hash", [False, True])
    @pytest.mark.parametrize("perform_removal", [True, False])
    @pytest.mark.parametrize(
        "num_buckets,jaccard_threshold,duplicate_docs",
        # Duplcated docs estimated from true_jaccard values
        [
            (5, 0.5, [[4, -1]]),
            (10, 0.39, [[4, -1], [1, 2]]),
            (15, 0.3, [[4, -1], [1, 2, 300]]),
        ],
    )
    def test_fuzzy_dedup(
        self,
        fuzzy_dedup_data,
        use_64_bit_hash,
        perform_removal,
        num_buckets,
        jaccard_threshold,
        duplicate_docs,
        tmpdir,
        gpu_client,
    ):
        print(gpu_client)
        # Dedup might fail when indices per partition do not start from 0
        fuzzy_dedup_data.df = fuzzy_dedup_data.df.reset_index(drop=True)
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=num_buckets,
            hashes_per_bucket=1,
            use_64_bit_hash=use_64_bit_hash,
            buckets_per_shuffle=3,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=jaccard_threshold,
        )
        fuzzy_duplicates = FuzzyDuplicates(
            config=config, perform_removal=perform_removal
        )
        result = fuzzy_duplicates(fuzzy_dedup_data)
        result_df = result.df.compute()
        if perform_removal:
            for duplicates in duplicate_docs:
                assert len(result_df[result_df["id"].isin(duplicates)]) == 1
        else:
            result_df = result_df[result_df.group.duplicated(keep=False)]
            result_df = result_df.groupby("group").id.agg(list)
            # Sort to maintain uniform ordering
            result_df = result_df.list.sort_values()
            result_df = result_df.sort_values()
            expected_df = cudf.Series(duplicate_docs, name="id")
            expected_df = expected_df.list.sort_values()
            expected_df = expected_df.sort_values()
            assert_eq(expected_df, result_df, check_index=False)

    def test_different_fields(self, fuzzy_dedup_data, tmpdir):
        fuzzy_dedup_data.df = fuzzy_dedup_data.df.reset_index(drop=True).rename(
            columns={"id": "col0", "text": "col1"}
        )
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="col0",
            text_field="col1",
            num_buckets=10,
            hashes_per_bucket=1,
            buckets_per_shuffle=5,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=0.39,
            char_ngrams=5,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        duplicates = fuzzy_duplicates.identify_duplicates(fuzzy_dedup_data)
        deduplicated_ds = fuzzy_duplicates.remove(fuzzy_dedup_data, duplicates)
        deduplicated_df = deduplicated_ds.df.compute()
        output_deduplicated_ids = set(deduplicated_df["col0"].to_arrow().to_pylist())
        assert len(deduplicated_df) == 3
        # From each of our groups we'll have atmost one document that is not duplicated
        assert (
            300 in output_deduplicated_ids
            and len({-1, 4}.intersection(output_deduplicated_ids)) == 1
            and len({1, 2}.intersection(output_deduplicated_ids)) == 1
        )

        # Drop non duplicated docs
        duplicates_df = duplicates.df.compute()
        duplicates_df = duplicates_df[duplicates_df.group.duplicated(keep=False)]
        duplicates_df = duplicates_df.groupby("group")["col0"].agg(list)
        # Sort to maintain uniform ordering
        duplicates_df = duplicates_df.list.sort_values()
        duplicates_df = duplicates_df.sort_values()

        duplicate_docs = [[4, -1], [1, 2]]
        expected_df = cudf.Series(duplicate_docs, name="col0")
        expected_df = expected_df.list.sort_values()
        expected_df = expected_df.sort_values()
        assert_eq(expected_df, duplicates_df, check_index=False)

    @pytest.mark.xfail
    def test_non_uniform_indices(
        self,
        tmpdir,
        gpu_client,
    ):
        print(gpu_client)
        # Dedup might fail when indices per partition do not start from 0
        df = cudf.DataFrame(
            {
                "id": [1, 2, 300, 4, -1],
                "text": [
                    "A test string",
                    "A different test string",
                    "A different object",
                    "The quick brown fox jumps over the lazy dog",
                    "The quick black cat jumps over the lazy dog",
                ],
            }
        )
        df = dask_cudf.from_cudf(df, 2)
        data = DocumentDataset(df)
        duplicate_docs = [[4, -1], [1, 2, 300]]
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=10,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=5,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=0.39,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        duplicates = fuzzy_duplicates.identify_duplicates(data)
        deduplicated_ds = fuzzy_duplicates.remove(fuzzy_dedup_data, duplicates)
        deduplicated_df = deduplicated_ds.df.compute()
        output_deduplicated_ids = set(deduplicated_df["col0"].to_arrow().to_pylist())
        assert len(deduplicated_df) == 2
        # From each of our groups we'll have atmost one document that is not duplicated
        assert (
            len({4, -1}.intersection(output_deduplicated_ids)) == 1
            and len({1, 2, 300}.intersection(output_deduplicated_ids)) == 1
        )

        duplicates_df = duplicates.df.compute()
        # Drop non duplicated docs
        duplicates_df = duplicates_df[duplicates_df.group.duplicated(keep=False)]
        duplicates_df = duplicates_df.groupby("group").id.agg(list)
        # Sort to maintain uniform ordering

        duplicates_df = duplicates_df.list.sort_values()
        duplicates_df = duplicates_df.sort_values()
        expected_df = cudf.Series(duplicate_docs, name="id")
        expected_df = expected_df.list.sort_values()
        expected_df = expected_df.sort_values()
        assert_eq(expected_df, duplicates_df, check_index=False)

    @pytest.mark.parametrize("num_anchors", [1, 3, 10])
    def test_num_anchors(self, large_fuzzy_dedup_data, num_anchors, tmpdir):
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=5,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=5,
            false_positive_check=True,
            num_anchors=num_anchors,
            jaccard_threshold=0.39,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        fuzzy_duplicates(large_fuzzy_dedup_data)
        anchor_docs_df_cols = dask_cudf.read_parquet(
            tmpdir / "anchor_docs_with_bk.parquet"
        ).columns
        assert all(f"anchor_{i}_id" in anchor_docs_df_cols for i in range(num_anchors))

    @pytest.mark.parametrize("use_64_bit_hash", [False, True])
    @pytest.mark.parametrize(
        "num_buckets,duplicate_docs",
        # Duplcated docs estimated from true_jaccard values
        [
            (10, [[4, -1], [1, 2, 300]]),
            (5, [[4, -1], [1, 2, 300]]),
        ],
    )
    def test_no_fp_check(
        self,
        fuzzy_dedup_data,
        use_64_bit_hash,
        num_buckets,
        duplicate_docs,
        tmpdir,
        gpu_client,
    ):
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=num_buckets,
            hashes_per_bucket=1,
            use_64_bit_hash=use_64_bit_hash,
            buckets_per_shuffle=3,
            false_positive_check=False,
            num_anchors=2,
            jaccard_threshold=0.39,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        result = fuzzy_duplicates.identify_duplicates(fuzzy_dedup_data)
        result_df = result.df.compute()
        # Drop non duplicated docs
        result_df = result_df[result_df.group.duplicated(keep=False)]
        result_df = result_df.groupby("group").id.agg(list)
        # Sort to maintain uniform ordering

        result_df = result_df.list.sort_values()
        result_df = result_df.sort_values()
        expected_df = cudf.Series(duplicate_docs, name="id")
        expected_df = expected_df.list.sort_values()
        expected_df = expected_df.sort_values()
        assert_eq(expected_df, result_df, check_index=False)

    def test_shuffle_fail_fuzzy_dedup_data(
        self,
        shuffle_fail_fuzzy_dedup_data,
        tmpdir,
        gpu_client,
    ):
        # Dedup might fail when indices per partition do not start from 0
        shuffle_fail_fuzzy_dedup_data.df = shuffle_fail_fuzzy_dedup_data.df.reset_index(
            drop=True
        )
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=10,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=5,
            false_positive_check=True,
            num_anchors=2,
            jaccard_threshold=0.39,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        result = fuzzy_duplicates.identify_duplicates(shuffle_fail_fuzzy_dedup_data)
        result_df = result.df.compute()
        # Drop non duplicated docs
        result_df = result_df[result_df.group.duplicated(keep=False)]
        result_df = result_df.groupby("group").id.agg(list)
        # Sort to maintain uniform ordering

        result_df = result_df.list.sort_values()
        result_df = result_df.sort_values()
        expected_df = cudf.Series([[1, 2]], name="id")
        expected_df = expected_df.list.sort_values()
        expected_df = expected_df.sort_values()
        assert_eq(expected_df, result_df, check_index=False)

    @pytest.mark.parametrize("false_positive_check", [True, False])
    def test_fuzzy_dedup_no_duplicates(
        self, no_duplicates_fuzzy_dedup_data, tmpdir, false_positive_check, gpu_client
    ):
        # Dedup might fail when indices per partition do not start from 0
        no_duplicates_fuzzy_dedup_data.df = (
            no_duplicates_fuzzy_dedup_data.df.reset_index(drop=True)
        )
        config = FuzzyDuplicatesConfig(
            cache_dir=tmpdir,
            id_field="id",
            text_field="text",
            seed=42,
            char_ngrams=5,
            num_buckets=10,
            hashes_per_bucket=1,
            use_64_bit_hash=False,
            buckets_per_shuffle=5,
            false_positive_check=false_positive_check,
            num_anchors=2,
            jaccard_threshold=0.39,
        )
        fuzzy_duplicates = FuzzyDuplicates(config=config, perform_removal=False)
        result = fuzzy_duplicates.identify_duplicates(no_duplicates_fuzzy_dedup_data)
        assert result is None


class TestFuzzyDuplicatesConfig:
    def test_bad_inputs(self, tmpdir):
        with pytest.raises(ValueError):
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir, num_anchors=0, false_positive_check=True
            )
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir, jaccard_threshold=1.2, false_positive_check=True
            )
            FuzzyDuplicatesConfig(cache_dir=tmpdir, buckets_per_shuffle=0)
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir, buckets_per_shuffle=2, num_buckets=1
            )
            FuzzyDuplicatesConfig(
                cache_dir=None, num_anchors=0, false_positive_check=True
            )
        with pytest.warns(
            UserWarning, match="Using a higher number of anchor docs might"
        ):
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir, num_anchors=3, false_positive_check=True
            )
        with pytest.warns(
            UserWarning, match="Using a small char_ngrams value might lead"
        ):
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir, char_ngrams=10, false_positive_check=False
            )
        with pytest.warns(
            UserWarning,
            match="Identifying false positives during the Minhash deduplication is computationally expensive",
        ):
            FuzzyDuplicatesConfig(cache_dir=tmpdir, false_positive_check=True)
        with pytest.warns(
            UserWarning,
            match="False positive check is disabled. Unused arguments",
        ):
            FuzzyDuplicatesConfig(
                cache_dir=tmpdir,
                false_positive_check=False,
                num_anchors=2,
                jaccard_threshold=0.8,
            )

    def test_from_yaml(self, tmpdir):
        yaml_params = {
            "cache_dir": "./",
            "num_anchors": 2,
            "jaccard_threshold": 0.8,
            "false_positive_check": False,
            "buckets_per_shuffle": 1,
            "char_ngrams": 20,
        }
        with open(tmpdir / "config.yaml", "w") as f:
            yaml.dump(yaml_params, f)
        config = FuzzyDuplicatesConfig.from_yaml(tmpdir / "config.yaml")
        for param in yaml_params:
            assert getattr(config, param) == yaml_params[param]


@pytest.mark.parametrize(
    "backend",
    [
        "pandas",
        pytest.param(
            "cudf",
            marks=pytest.mark.gpu,
        ),
    ],
)
def test_extract_partitioning_index(backend):

    def add_partition_info(df, partition_info=None):
        if partition_info is None:
            df["file_id"] = -1
        else:
            df["file_id"] = partition_info["number"]
        return df

    with config.set({"dataframe.backend": backend}):

        # Create a random `unshuffled` DataFrame with a
        # "part_id" column to be used as the shuffle index
        npartitions_left = 7
        unshuffled = dd.from_dict(
            {"part_id": np.random.randint(25, size=1000, dtype="int32")},
            npartitions=npartitions_left,
        )

        # Create a `bk_mapping` DataFrame that defines
        # the "correct" mapping beween "part_id" and
        # the destination partition ("file_id")
        npartitions_right = 5
        bk_mapping = (
            dd.from_dict(
                {"part_id": np.arange(25, dtype="int32")},
                npartitions=npartitions_right,
            )
            .shuffle("part_id")
            .map_partitions(add_partition_info)
            .compute()
        )

    # Use `extract_partitioning_index` to calculate
    # the partitioning index and assign it as a new
    # "_partitions" column
    result, _ = extract_partitioning_index(
        unshuffled,
        "part_id",
        bk_mapping,
        npartitions_right,
        npartitions_right,
    )

    # Rename the "_partitions" column, shuffle by "part_id",
    # and then assign a "file_id" column to reflect the final
    # partition of each row
    check = (
        result.rename(columns={"_partitions": "expected_file_id"})
        .shuffle(
            "part_id",
            npartitions=npartitions_right,
        )
        .map_partitions(add_partition_info)
        .compute()
    )

    # Check that the real and expected partitions match
    assert (check["file_id"] == check["expected_file_id"]).all()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from typing import Dict

import requests

from nemo_curator.download.doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
)


class EmailsDownloader(DocumentDownloader):
    def __init__(self, download_dir: str):
        super().__init__()

        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)

        self._download_dir = download_dir
        print("Download directory: ", self._download_dir)

    def download(self, url: str) -> str:
        filename = os.path.basename(url)
        output_file = os.path.join(self._download_dir, filename)

        if os.path.exists(output_file):
            print(f"File '{output_file}' already exists, skipping download.")
            return output_file

        print(f"Downloading Enron emails dataset from '{url}'...")
        response = requests.get(url)

        with open(output_file, "wb") as file:
            file.write(response.content)

        return output_file


class EmailsIterator(DocumentIterator):

    def __init__(self):
        super().__init__()
        self._counter = -1
        self._extractor = EmailsExtractor()
        # The regular expression pattern to extract each email.
        self._pattern = re.compile(r"\"<s>.*?<s>\"", re.DOTALL)

    def iterate(self, file_path):
        self._counter = -1
        file_name = os.path.basename(file_path)

        with open(file_path, "r", encoding="utf-8") as file:
            lines = file.readlines()

        # Ignore the first line which contains the header.
        file_content = "".join(lines[1:])
        # Find all the emails in the file.
        it = self._pattern.finditer(file_content)

        for email in it:
            self._counter += 1
            content = email.group().strip('"').strip()
            meta = {
                "file_name": file_name,
                "id": f"email-{self._counter}",
            }
            extracted_content = self._extractor.extract(content)

            # Skip if no content extracted
            if not extracted_content:
                continue

            record = {**meta, **extracted_content}
            yield record


class EmailsExtractor(DocumentExtractor):
    def __init__(self):
        super().__init__()
        # The regular expression pattern to extract subject/body/label into groups.
        self._pattern = re.compile(
            r"Subject:: (.*?)\nBody:: (.*?)\n.*\[/INST\] (.*?) <s>", re.DOTALL
        )

    def extract(self, content: str) -> Dict[str, str]:
        matches = self._pattern.findall(content)

        if not matches:
            return None

        matches = matches[0]

        return {
            "subject": matches[0].strip(),
            "body": matches[1].strip(),
            "category": matches[2].strip(),
        }



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from datetime import datetime

from nemo_curator.log import create_logger
from nemo_curator.modules import SemanticClusterLevelDedup
from nemo_curator.modules.config import SemDedupConfig
from nemo_curator.utils.distributed_utils import get_client
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    semdedup_config = SemDedupConfig.from_yaml(args.config_file)
    client = get_client(**ArgumentHelper.parse_client_args(args))

    root = semdedup_config.cache_dir
    save_loc = semdedup_config.clustering_save_loc
    client = get_client(**ArgumentHelper.parse_client_args(args))

    logger = create_logger(
        rank=0,
        log_file=os.path.join(root, save_loc, "extract_dedup_data.log"),
        name="logger-extract-dedup-data",
        log_level=logging.INFO,
        stdout=True,
    )

    dt1 = datetime.now()
    logger.info(f"Start: {dt1}")
    cache_dir = semdedup_config.cache_dir

    semantic_dedup = SemanticClusterLevelDedup(
        n_clusters=semdedup_config.n_clusters,
        emb_by_clust_dir=os.path.join(
            cache_dir, semdedup_config.clustering_save_loc, "embs_by_nearest_center"
        ),
        id_column=args.id_column,
        which_to_keep=semdedup_config.which_to_keep,
        sim_metric=semdedup_config.sim_metric,
        batched_cosine_similarity=semdedup_config.batched_cosine_similarity,
        output_dir=os.path.join(
            semdedup_config.cache_dir, semdedup_config.clustering_save_loc
        ),
        embedding_column=semdedup_config.embedding_column,
        logger=logger,
    )

    semantic_dedup.compute_semantic_match_dfs()
    dedup_id_dataset = semantic_dedup.extract_dedup_data(
        eps_to_extract=semdedup_config.eps_to_extract
    )
    print(dedup_id_dataset.df.head(10))

    dt2 = datetime.now()
    logger.info(f"End: {dt2}")
    elapse = (dt2 - dt1).total_seconds() / 60
    logger.info(f"elapse: {elapse}")

    client.cancel(client.futures, force=True)
    client.close()


def attach_args():
    parser = ArgumentHelper.parse_semdedup_args(
        description=(
            "Extracts deduplicated data from the clustered embeddings of a collection of documents. "
            "This script requires that embeddings and clustering have been performed "
            "earlier using semdedup_extract_embeddings and semdedup_cluster_embeddings."
            "Input arguments include: "
            "--id-column for the the identifier in the dataset, "
            "--id-column-type for the data type of ID column, "
            "--config-file for the path to the semantic deduplication configuration file. "
            "Important configuration parameters include:"
            " cache_dir for the directory to store cache"
            " which_to_keep for specifying which duplicates to keep,"
            " sim_metric for the similarity metric for deduplication,"
            " and eps_to_extract for the epsilon value to extract deduplicated data."
        ),
    )
    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import time
from typing import Optional, Union

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.modules.base import BaseDeduplicationModule
from nemo_curator.modules.config import FuzzyDuplicatesConfig
from nemo_curator.modules.fuzzy_dedup._mapbuckets import _MapBuckets
from nemo_curator.modules.fuzzy_dedup._shuffle import _Shuffle
from nemo_curator.modules.fuzzy_dedup.bucketstoedges import BucketsToEdges
from nemo_curator.modules.fuzzy_dedup.connectedcomponents import ConnectedComponents
from nemo_curator.modules.fuzzy_dedup.jaccardsimilarity import JaccardSimilarity
from nemo_curator.modules.fuzzy_dedup.lsh import LSH
from nemo_curator.modules.fuzzy_dedup.minhash import MinHash
from nemo_curator.modules.meta import Sequential
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix
from nemo_curator.utils.duplicates_removal import remove_duplicates


class FuzzyDuplicates(BaseDeduplicationModule):
    def __init__(
        self,
        config: FuzzyDuplicatesConfig,
        logger: Union[logging.LoggerAdapter, str] = "./",
        perform_removal: bool = False,
    ):
        """
        Parameters
        ----------
        config: FuzzyDuplicatesConfig,
            Config options for finding FuzzyDuplicates
        logger: Existing logger to log to, or a path to a log directory.
        perform_removal: Whether to remove duplicates from the dataset.
            Default is False.
        Returns
        -------
        DocumentDataset containing IDs of all documents and the corresponding duplicate group
        they belong to. Documents in the same group are near duplicates.
        """
        super().__init__(
            id_field=config.id_field,
            text_field=config.text_field,
            input_backend="cudf",
            logger=logger,
            perform_removal=perform_removal,
            profile_dir=config.profile_dir,
            cache_dir=config.cache_dir,
        )

        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "FuzzyDuplicates.log"),
                name="FuzzyDuplicates",
            )
        else:
            self._logger = logger

        self.config = config
        self.minhash = MinHash(
            seed=self.config.seed,
            num_hashes=self.config.num_hashes,
            char_ngrams=self.config.char_ngrams,
            use_64bit_hash=self.config.use_64_bit_hash,
            logger=self._logger,
            id_field=self.config.id_field,
            text_field=self.config.text_field,
            profile_dir=self.config.profile_dir,
            cache_dir=self.config.cache_dir,
        )
        self.lsh = LSH(
            cache_dir=self.config.cache_dir,
            num_hashes=self.config.num_hashes,
            num_buckets=self.config.num_buckets,
            buckets_per_shuffle=self.config.buckets_per_shuffle,
            false_positive_check=self.config.false_positive_check,
            logger=self._logger,
            id_fields=[self.config.id_field],
            profile_dir=self.config.profile_dir,
        )

        if self.config.false_positive_check:
            self.map_buckets = _MapBuckets(
                id_fields=[self.config.id_field],
                text_field=self.config.text_field,
                logger=self._logger,
                num_anchors=self.config.num_anchors,
            )
            self.jaccard_shuffle = _Shuffle(
                id_fields=[self.config.id_field],
                text_field=self.config.text_field,
                logger=self._logger,
                profile_dir=self.config.profile_dir,
            )
            self.jaccard_compute = JaccardSimilarity(
                id_field=self.config.id_field,
                text_field=self.config.text_field,
                ngram_width=self.config.char_ngrams,
                anchor_id_fields=[
                    f"anchor_{i}_{self.config.id_field}"
                    for i in range(self.config.num_anchors)
                ],
            )
        else:
            self.buckets_to_edges = BucketsToEdges(
                cache_dir=self.config.cache_dir,
                id_fields=self.config.id_field,
                logger=self._logger,
                profile_dir=self.config.profile_dir,
            )

        jaccard_pairs_fname = (
            "jaccard_similarity_results.parquet"
            if self.config.false_positive_check
            else "_edges.parquet"
        )
        self.connected_components = ConnectedComponents(
            cache_dir=self.config.cache_dir,
            jaccard_pairs_path=os.path.join(self.config.cache_dir, jaccard_pairs_fname),
            id_column=self.config.id_field,
            jaccard_threshold=self.config.jaccard_threshold,
            logger=self._logger,
            profile_dir=self.config.profile_dir,
        )

    def identify_duplicates(
        self, dataset: DocumentDataset
    ) -> Optional[DocumentDataset]:
        """
        Parameters
        ----------
        dataset: DocumentDataset
            The input datset to compute FuzzyDuplicates. Must contain a text and unique id field.

        Returns
        -------
        DocumentDataset containing IDs of all documents and the corresponding duplicate group
        they belong to. Documents in the same group are near duplicates.
        """

        # Minhash + LSH
        stage_num = 1
        print(f"Stage {stage_num}: Starting Minhash + LSH computation")
        minhashLSH = Sequential([self.minhash, self.lsh])
        buckets_df = minhashLSH(dataset)
        print(f"Stage {stage_num}: Minhash + LSH complete!")
        if buckets_df is None:
            print(
                f"Stage {stage_num}: No potential duplicate documents found during LSH"
            )
            return None
        stage_num += 1

        if self.config.false_positive_check:
            # Map buckets to lower cardinality distribution
            print(f"Stage {stage_num} (False Positive Check): Starting Map_Buckets")
            t0 = time.time()
            mapped_buckets_w_anchors_path = os.path.join(
                self.config.cache_dir, "anchor_docs_with_bk.parquet"
            )
            with performance_report_if_with_ts_suffix(
                self.config.profile_dir,
                "map_buckets",
            ):
                ddf_mapped_buckets_w_anchors = (
                    self.map_buckets.map_buckets_with_anchors(
                        documents_df=dataset.df, buckets_df=buckets_df.df
                    )
                )
                ddf_mapped_buckets_w_anchors.to_parquet(
                    mapped_buckets_w_anchors_path, write_index=False, overwrite=True
                )
            self._logger.info(
                f"Time taken for Map_buckets : {time.time() - t0}s and output written at {mapped_buckets_w_anchors_path}"
            )

            print(f"Stage {stage_num} (False Postive Check): Map_Buckets Complete!")
            stage_num += 1

            # Shuffle documents based on mapped buckets
            print(f"Stage {stage_num} (False Postive Check): Shuffle docs")
            shuffled_docs_path = os.path.join(
                self.config.cache_dir, "shuffled_docs.parquet"
            )
            self.jaccard_shuffle.shuffle_docs_on_buckets(
                documents_df=dataset.df,
                bucket_w_anchors_path=mapped_buckets_w_anchors_path,
                output_shuffled_docs_path=shuffled_docs_path,
                bucket_mapping_df_blocksize=self.config.bucket_mapping_blocksize,
                parts_per_worker=self.config.parts_per_worker,
                bucket_parts_per_worker=self.config.bucket_parts_per_worker,
            )
            print(f"Stage {stage_num} (False Postive Check): Shuffle docs complete!")
            stage_num += 1

            # jaccard comparision within buckets
            print(
                f"Stage {stage_num} (False Postive Check): Jaccard Similarity in Buckets"
            )
            jaccard_pairs_path = os.path.join(
                self.config.cache_dir, "jaccard_similarity_results.parquet"
            )
            t0 = time.time()
            with performance_report_if_with_ts_suffix(
                self.config.profile_dir,
                "jaccard-similarity",
            ):
                jaccard_pairs_df = self.jaccard_compute.jaccard_compute(
                    shuffled_docs_path=shuffled_docs_path
                )
                jaccard_pairs_df.to_parquet(
                    jaccard_pairs_path,
                    write_index=False,
                    write_metadata_file=False,
                    overwrite=True,
                )
                self._logger.info(
                    f"Time taken for Jaccard Similarity = {time.time()-t0}s and output written at {jaccard_pairs_path}"
                )

            print(
                f"Stage {stage_num} (False Postive Check): Jaccard Similarity in Buckets Complete!"
            )
            stage_num += 1

        else:
            # Map buckets to lower cardinality distribution
            print(f"Stage {stage_num}: Starting LSH Buckets to Graph Edgelist")
            self.buckets_to_edges(buckets_df)
            print(
                f"Stage {stage_num}: Starting LSH Buckets to Graph Edgelist Complete!"
            )
            stage_num += 1

        # Connected components across buckets
        print(f"Stage {stage_num}: Connected Components across buckets")
        cc_path = os.path.join(self.config.cache_dir, "connected_components.parquet")
        self.connected_components.cc_workflow(cc_path)
        print(f"Stage {stage_num}: Connected Components across buckets complete!")
        stage_num += 1

        return DocumentDataset.read_parquet(
            cc_path,
            backend="cudf",
            # We read with files_per_partition=1 so that groups are read in whole (and do not exist across partitions)
            files_per_partition=1,
            blocksize=None,
        )

    def remove(
        self, dataset: DocumentDataset, duplicates_to_remove: Optional[DocumentDataset]
    ) -> DocumentDataset:
        """
        Remove fuzzy duplicates from a given DocumentDataset
        Parameters
        ----------
        dataset: DocumentDataset
          The input dataset from which to remove fuzzy duplicates
        duplicates_to_remove: DocumentDataset
          The dataset containing IDs of the fuzzy duplicates to remove
        Returns
        -------
        DocumentDataset containing only non-duplicate documents
        """

        if not duplicates_to_remove:
            print("No fuzzy duplicates to remove, returning original dataset")
            return dataset

        result = remove_duplicates(
            left=dataset.df,
            duplicates=duplicates_to_remove.df,
            id_field=self.config.id_field,
            group_field="group",
        )
        return DocumentDataset(result)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Dict

from nemo_curator.utils.import_utils import safe_import

run = safe_import("nemo_run")


@dataclass
class SlurmJobConfig:
    """
    Configuration for running a NeMo Curator script on a Slurm cluster using
    NeMo Run

    Args:
        job_dir: The base directory where all the files related to setting up
            the Dask cluster for NeMo Curator will be written
        container_entrypoint: A path to the container-entrypoint.sh script
            on the cluster. container-entrypoint.sh is found in the repo
            here: https://github.com/NVIDIA/NeMo-Curator/blob/main/examples/slurm/container-entrypoint.sh
        script_command: The NeMo Curator CLI tool to run. Pass any additional arguments
            needed directly in this string.
        device: The type of script that will be running, and therefore the type
            of Dask cluster that will be created. Must be either "cpu" or "gpu".
        interface: The network interface the Dask cluster will communicate over.
            Use nemo_curator.get_network_interfaces() to get a list of available ones.
        protocol: The networking protocol to use. Can be either "tcp" or "ucx".
            Setting to "ucx" is recommended for GPU jobs if your cluster supports it.
        cpu_worker_memory_limit: The maximum memory per process that a Dask worker can use.
            "5GB" or "5000M" are examples. "0" means no limit.
        rapids_no_initialize: Will delay or disable the CUDA context creation of RAPIDS libraries,
            allowing for improved compatibility with UCX-enabled clusters and preventing runtime warnings.
        cudf_spill: Enables automatic spilling (and “unspilling”) of buffers from device to host to
            enable out-of-memory computation, i.e., computing on objects that occupy more memory
            than is available on the GPU.
        rmm_scheduler_pool_size: Sets a small pool of GPU memory for message transfers when
            the scheduler is using ucx
        rmm_worker_pool_size: The amount of GPU memory each GPU worker process may use.
            Recommended to set at 80-90% of available GPU memory. 72GiB is good for A100/H100
        libcudf_cufile_policy: Allows reading/writing directly from storage to GPU.
    """

    job_dir: str
    container_entrypoint: str
    script_command: str
    device: str = "cpu"
    interface: str = "eth0"
    protocol: str = "tcp"
    cpu_worker_memory_limit: str = "0"
    rapids_no_initialize: str = "1"
    cudf_spill: str = "1"
    rmm_scheduler_pool_size: str = "1GB"
    rmm_worker_pool_size: str = "72GiB"
    libcudf_cufile_policy: str = "OFF"

    def to_script(self, add_scheduler_file: bool = True, add_device: bool = True):
        """
        Converts to a script object executable by NeMo Run
        Args:
            add_scheduler_file: Automatically appends a '--scheduler-file' argument to the
                script_command where the value is job_dir/logs/scheduler.json. All
                scripts included in NeMo Curator accept and require this argument to scale
                properly on Slurm clusters.
            add_device: Automatically appends a '--device' argument to the script_command
                where the value is the member variable of device. All scripts included in
                NeMo Curator accept and require this argument.
        Returns:
            A NeMo Run Script that will intialize a Dask cluster, and run the specified command.
            It is designed to be executed on a Slurm cluster
        """
        env_vars = self._build_env_vars()

        if add_scheduler_file:
            env_vars[
                "SCRIPT_COMMAND"
            ] += f" --scheduler-file={env_vars['SCHEDULER_FILE']}"
        if add_device:
            env_vars["SCRIPT_COMMAND"] += f" --device={env_vars['DEVICE']}"

        # Surround the command in quotes so the variable gets set properly
        env_vars["SCRIPT_COMMAND"] = f"\"{env_vars['SCRIPT_COMMAND']}\""

        return run.Script(path=self.container_entrypoint, env=env_vars)

    def _build_env_vars(self) -> Dict[str, str]:
        env_vars = vars(self)
        # Convert to uppercase to match container_entrypoint.sh
        env_vars = {key.upper(): val for key, val in env_vars.items()}

        env_vars["LOGDIR"] = f"{self.job_dir}/logs"
        env_vars["PROFILESDIR"] = f"{self.job_dir}/profiles"
        env_vars["SCHEDULER_FILE"] = f"{env_vars['LOGDIR']}/scheduler.json"
        env_vars["SCHEDULER_LOG"] = f"{env_vars['LOGDIR']}/scheduler.log"
        env_vars["DONE_MARKER"] = f"{env_vars['LOGDIR']}/done.txt"

        return env_vars



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

from nemo_curator.download.doc_builder import batch_download, download_and_extract
from nemo_curator.utils.config_utils import build_downloader
from nemo_curator.utils.distributed_utils import get_client, write_to_disk
from nemo_curator.utils.file_utils import (
    expand_outdir_and_mkdir,
    get_all_files_paths_under,
)
from nemo_curator.utils.script_utils import ArgumentHelper


def read_urls(file_path):
    with open(file_path, "r") as fp:
        urls = fp.readlines()
    return [url.strip() for url in urls]


def main(args):
    client = get_client(**ArgumentHelper.parse_client_args(args))

    if args.input_url_file:
        urls = read_urls(args.input_url_file)
        outdir = os.path.abspath(os.path.expanduser(args.output_json_dir))
        output_paths = list(
            map(lambda url: os.path.join(outdir, url.split("/")[-1] + ".jsonl"), urls)
        )
    elif args.input_data_dir:
        # If input_data_dir is specified, we operate in extraction only mode.
        urls = get_all_files_paths_under(args.input_data_dir)
        output_paths = urls
    else:
        raise ValueError(
            "One of --input-url-file or --input-data-dir must be specified"
        )

    expand_outdir_and_mkdir(args.output_json_dir)
    if args.output_download_dir:
        raw_download_dir = args.output_download_dir
    else:
        raw_download_dir = os.path.join(args.output_json_dir, "downloads")

    downloader, iterator, extractor, output_format = build_downloader(
        args.builder_config_file, default_download_dir=raw_download_dir
    )

    if args.download_only:
        output_paths = batch_download(urls, downloader)
        print(f"{len(output_paths)} were downloaded")
        return

    dataset = download_and_extract(
        urls,
        output_paths,
        downloader,
        iterator,
        extractor,
        output_format,
        keep_raw_download=args.keep_downloaded_files,
        force_download=args.overwrite_existing_json,
        input_meta=args.input_meta,
        record_limit=args.record_limit,
    )

    # Sample to trigger the dask computation
    write_to_disk(
        dataset.df,
        args.output_json_dir,
        write_to_filename=True,
        output_type="jsonl",
    )


def attach_args(
    parser=argparse.ArgumentParser(
        """
Takes an input list of URLs, downloads the data,
and then extracts the text from the downloaded data. Using
the --builder-config-file argument, users must provide a YAML file
that points to implementations of the downloader, iterator, and extractor
classes that will be used to construct the documents that will "
"make up the dataset. Examples of these config files for the "
"Common Crawl, Wikipedia and ArXiv datasets can be found in the root "
"config directory of this repository.

In the case that users have data that have been pre-downloaded,
this utility can also be used for "extraction-only" purposes.
For this scenario, users should either provide a valid path
to the --input-data-dir/--input-local-data-dir directories
(depending on if the data are globally or locally available to each
MPI rank). Additionally, the downloader class should be implemented
such that it simply returns the pre-downloaded file.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    argumentHelper.add_arg_input_data_dir(help="Path to input data directory.")
    argumentHelper.add_arg_input_meta()
    argumentHelper.add_distributed_args()
    parser.add_argument(
        "--builder-config-file",
        type=str,
        required=True,
        help="YAML file that contains paths to implementations of a downloader, "
        "iterator, and extractor that will be used in this program "
        "to build the documents that make up the output dataset.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "download-only",
        help="Specify this flag if you desire to only download the data "
        "files and not extract text from the downloaded files.",
    )
    parser.add_argument(
        "--input-url-file",
        type=str,
        default=None,
        help="Input directory consisting of .jsonl files that are accessible "
        "to all nodes. Use this for a distributed file system.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "keep-downloaded-files",
        help="If this flag is set to true, the downloaded data files "
        "will be kept on disk and not removed after extraction.",
    )
    parser.add_argument(
        "--output-download-dir",
        type=str,
        default=None,
        help="The directory to where data files will be written "
        'in "download-only" mode. Specify this argument only when '
        "the --download-only flag is specified.",
    )
    parser.add_argument(
        "--output-json-dir",
        type=str,
        default=None,
        help="Output directory to store the extracted text in JSONL files.",
    )
    parser.add_argument(
        "--record-limit",
        type=int,
        default=None,
        help="Limit the number of records to extract from each file.",
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "overwrite-existing-json",
        help="If this flag is specified, then the JSON data will be "
        "overwritten if downloading from the the same file.",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from nemo_curator.utils.download_utils import get_common_crawl_urls
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    urls = get_common_crawl_urls(
        args.starting_snapshot,
        args.ending_snapshot,
        data_domain_prefix=args.cc_data_domain_prefix,
        index_prefix=args.cc_index_prefix,
        news=args.cc_news,
    )

    with open(args.output_warc_url_file, "w") as fp:
        for url in urls:
            fp.write(url)
            fp.write("\n")


def attach_args(
    parser=argparse.ArgumentParser(
        """
Pulls URLs of WARC files stored within the Common Crawl data repository
and writes them to file so that they can be used to subsequently
download the WARC files.
""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
):
    argumentHelper = ArgumentHelper(parser)

    parser.add_argument(
        "--cc-data-domain-prefix",
        type=str,
        default="https://data.commoncrawl.org",
        help="The prefix that will be prepended to each WARC "
        "file to create the URL. By default this value is "
        ' "https://data.commoncrawl.org".',
    )
    parser.add_argument(
        "--ending-snapshot",
        type=str,
        default="2020-50",
        help="The last snapshot for which WARC URLs will be retrieved. "
        "Snapshots must be specified by YYYY-WeekNumber "
        '(e.g., "2020-50" or "2021-04").',
    )
    parser.add_argument(
        "--cc-index-prefix",
        type=str,
        default="https://index.commoncrawl.org",
        help="The prefix of the URL to the Common Crawl index. "
        'By default this value is "https://index.commoncrawl.org".',
    )
    ArgumentHelper.attach_bool_arg(
        parser,
        "cc-news",
        help="Specify --cc-news in order to download WARC URLs for "
        "the CC-NEWS dataset instead of the CC-MAIN datasets. If this "
        "is specified, then it is assumed that the format for the start "
        'and end snapshots is "YYYY-MM" (Year-Month). All WARC URLs between '
        "the specified years and months will be downloaded.",
    )
    parser.add_argument(
        "--output-warc-url-file",
        type=str,
        default=None,
        required=True,
        help="The output file to which the WARC URLs will be written.",
    )
    parser.add_argument(
        "--starting-snapshot",
        type=str,
        default="2020-50",
        help="The starting snapshot to download. All WARC URLs will be written "
        "between the dates specified by --starting-snapshot "
        "and --ending-snapshot. Snapshots must be specified by YYYY-WeekNumber "
        '(e.g., "2020-50" or "2021-04"). For the CC-NEWS dataset, '
        "(specified with the --cc-news flag), this changes to "
        "Year-Month (YYYY-MM).",
    )

    return parser


def console_script():
    main(attach_args().parse_args())



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def convert_str_id_to_int(df, id_column="id"):
    """
    Converts the legacy id format "dataset_name-0000034"
    type of ID into 2 int based ID's
    """
    dx = df[id_column].str.rsplit("-", n=1, expand=True)
    df["doc_id"] = dx[1].astype("int64").values
    df["dataset_id"] = dx[0].hash_values()
    df.drop(columns=[id_column], inplace=True)
    return df


def int_ids_to_str(df, id_column="id"):
    """
    Converts int id's generated via `convert_str_id_to_int`
    back to a string ID
    """
    df[id_column] = df["dataset_id"].astype(str) + "-" + df["doc_id"].astype(str)
    df.drop(columns=["dataset_id", "doc_id"], inplace=True)

    if "anchor_0_dataset_id" in df.columns:
        df[f"anchor_0_{id_column}"] = (
            df["anchor_0_dataset_id"].astype(str)
            + "-"
            + df["anchor_0_doc_id"].astype(str)
        )
        df.drop(columns=["anchor_0_dataset_id", "anchor_0_doc_id"], inplace=True)

    if "anchor_1_dataset_id" in df.columns:
        df[f"anchor_1_{id_column}"] = (
            df["anchor_1_dataset_id"].astype(str)
            + "-"
            + df["anchor_1_doc_id"].astype(str)
        )
        df.drop(columns=["anchor_1_dataset_id", "anchor_1_doc_id"], inplace=True)
    return df



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.modules import EmbeddingCreator
from nemo_curator.modules.config import SemDedupConfig
from nemo_curator.utils.distributed_utils import get_client, read_data
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir, get_remaining_files
from nemo_curator.utils.script_utils import ArgumentHelper


def main(args):
    semdedup_config = SemDedupConfig.from_yaml(args.config_file)
    client = get_client(**ArgumentHelper.parse_client_args(args))
    expand_outdir_and_mkdir(semdedup_config.cache_dir)
    logger = create_logger(
        rank=0,
        name="logger-compute-embeddings",
        log_file=os.path.join(semdedup_config.cache_dir, "compute_embeddings.log"),
        log_level=logging.INFO,
        stdout=True,
    )

    output_data_dir = os.path.join(
        semdedup_config.cache_dir, semdedup_config.embeddings_save_loc
    )

    # Sometimes JSONL files are stored as .json
    # So to handle that case we can pass the input_file_extension
    if args.input_file_extension is not None:
        input_file_extension = args.input_file_extension
    else:
        input_file_extension = args.input_file_type
    print("input_file_extension", input_file_extension)

    st = time.time()
    input_files = get_remaining_files(
        input_file_path=args.input_data_dir,
        output_file_path=output_data_dir,
        input_file_type=input_file_extension,
        output_file_type="parquet",
        num_files=semdedup_config.num_files,
    )

    logger.info(f"Processing {len(input_files)} files")
    if len(input_files) == 0:
        logger.info("No files to process")
        return

    ddf = read_data(
        input_files=input_files,
        file_type=args.input_file_type,
        add_filename=True,
    )
    ddf = ddf.reset_index(drop=True)
    dataset = DocumentDataset(ddf)

    # Can repartition here if needed
    # ddf = ddf.repartition(partition_size="64MB")
    embedding_creator = EmbeddingCreator(
        embedding_model_name_or_path=semdedup_config.embedding_model_name_or_path,
        embedding_batch_size=semdedup_config.embedding_batch_size,
        embedding_output_dir=os.path.join(
            semdedup_config.cache_dir, semdedup_config.embeddings_save_loc
        ),
        embedding_max_mem_gb=semdedup_config.embedding_max_mem_gb,
        embedding_pooling_strategy=semdedup_config.embedding_pooling_strategy,
        input_column=args.input_text_field,
        embedding_column=semdedup_config.embedding_column,
        write_embeddings_to_disk=semdedup_config.write_embeddings_to_disk,
        write_to_filename=semdedup_config.write_to_filename,
        logger=logger,
    )

    embedding_dataset = embedding_creator(dataset=dataset)
    print(embedding_dataset.df.head())
    logger.info(f"Time taken: {time.time() - st}")
    client.cancel(client.futures, force=True)
    client.close()


def attach_args():
    parser = ArgumentHelper.parse_semdedup_args(
        description=(
            "Computes the embeddings of a collection of documents using the specified model. "
            'The model is specified in the configuration file using embedding_model_name_or_path (e.g. "sentence-transformers/paraphrase-MiniLM-L6-v2"). '
            "The embeddings are saved in the specified cache directory under the embeddings_save_loc directory. "
            "Input arguments include: "
            "--input-data-dir for the directory containing input data files, "
            '--input-file-type for the type of input files (e.g., "json", "csv"), '
            '--input-file-extension for specifying the file extension of input files (e.g., ".jsonl"), '
            "--input-text-field for the field in the input files containing the text data to be embedded, "
            "--config-file for the path to the semantic deduplication configuration file. "
            "Important configuration parameters include: "
            " cache_dir for the directory to store cache"
            " num_files for the number of files to process (default is -1, meaning all files),"
            " embeddings_save_loc for the location to save embeddings,"
            " embedding_model_name_or_path for the model name or path for embeddings,"
            " embedding_batch_size for the batch size for processing embeddings."
        ),
    )
    return parser


def console_script():
    main(attach_args().parse_args())


if __name__ == "__main__":
    main(attach_args().parse_args())



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

import pandas as pd
from docbuilder import (
    ArxivDownloader,
    ArxivExtractor,
    ArxivIterator,
    GitHubDownloader,
    GitHubExtractor,
    GitHubIterator,
    WikitxtDownloader,
    WikitxtExtractor,
    WikitxtIterator,
)

from nemo_curator.download.doc_builder import download_and_extract


def download_wikipedia_sources(
    source_links_file: Optional[str] = None,
    output_dir: Optional[str] = None,
    limit: Optional[int] = None,
) -> str:
    """
    Downloads Wikipedia sources based on the provided source links file.

    Args:
        source_links_file: Path to the file containing the source links. If not provided,
            a default file path (./sources/wikipedia_urls.jsonl) will be used.
        output_dir: Directory where the downloaded sources will be saved. If not provided,
            a default directory path (./data/raw/wikipedia) will be used.
        limit: Maximum number of sources to download. If provided, only the first `limit`
            sources will be downloaded.

    Returns:
        str: The path to the output directory where the downloaded sources are saved.
    """

    if source_links_file is None:
        source_links_file = os.path.join(
            os.path.dirname(__file__), "sources", "wikipedia_urls.jsonl"
        )

    if not os.path.exists(source_links_file):
        raise FileNotFoundError(f"File '{source_links_file}' not found.")

    if output_dir is None:
        output_dir = os.path.join(os.path.dirname(__file__), "data", "raw", "wikipedia")

    os.makedirs(output_dir, exist_ok=True)
    # Read the source links from the provided file
    urls = pd.read_json(path_or_buf=source_links_file, lines=True)
    urls = urls[0].tolist()

    if limit:
        urls = urls[:limit]
    output_format = {
        "text": str,
        "id": str,
        "file_extension": str,
        "file_type": str,
        "category": str,
        "line_count": int,
        "size_in_bytes": int,
        "path": str,
        "file_name": str,
    }

    downloader = WikitxtDownloader(output_dir)
    iterator = WikitxtIterator()
    extractor = WikitxtExtractor()

    dataset = download_and_extract(
        urls=urls,
        output_paths=[os.path.join(output_dir, os.path.basename(url)) for url in urls],
        downloader=downloader,
        iterator=iterator,
        extractor=extractor,
        output_format=output_format,
    )

    dataset.to_json(output_dir, write_to_filename="file_name")
    return output_dir


def download_github_sources(
    source_links_file: Optional[str] = None,
    output_dir: Optional[str] = None,
    limit: Optional[int] = None,
):
    """
    Downloads GitHub sources specified in a file and extracts them.

    Args:
        source_links_file: Path to the file containing the GitHub source links.
        output_dir: Directory where the downloaded and extracted files will be stored.
        limit: Maximum number of GitHub sources to download and extract.

    Raises:
        FileNotFoundError: If the source_links_file does not exist.

    Returns:
        str: Path to the output directory where the downloaded and extracted files are stored.
    """

    if source_links_file is None:
        source_links_file = os.path.join(
            os.path.dirname(__file__), "sources", "github_repos.jsonl"
        )

    if not os.path.exists(source_links_file):
        raise FileNotFoundError(f"File '{source_links_file}' not found.")

    urls = pd.read_json(path_or_buf=source_links_file, lines=True)
    urls = urls[0].tolist()

    if limit:
        urls = urls[:limit]

    if output_dir is None:
        output_dir = os.path.join(os.path.dirname(__file__), "data", "raw", "github")

    os.makedirs(output_dir, exist_ok=True)
    output_jsonl_dir = os.path.join(output_dir, "jsonl")
    os.makedirs(output_jsonl_dir, exist_ok=True)

    downloader = GitHubDownloader(output_dir)
    iterator = GitHubIterator()
    extractor = GitHubExtractor()

    output_format = {
        "text": str,
        "id": str,
        "file_extension": str,
        "file_type": str,
        "category": str,
        "line_count": int,
        "size_in_bytes": int,
        "path": str,
        "file_name": str,
    }

    dataset = download_and_extract(
        urls=urls,
        output_paths=[
            os.path.join(output_jsonl_dir, os.path.basename(url)) for url in urls
        ],
        downloader=downloader,
        iterator=iterator,
        extractor=extractor,
        output_format=output_format,
        keep_raw_download=True,
    )

    dataset.to_json(output_jsonl_dir, write_to_filename="file_name")
    return output_jsonl_dir


def download_pdf_sources(
    source_links_file: Optional[str] = None,
    output_dir: Optional[str] = None,
    limit: Optional[int] = None,
):
    """
    Downloads Arxiv Pdf sources specified in a file and extracts them.

    Args:
        source_links_file: Path to the file containing the Arxiv PDF source links.
        output_dir: Directory where the downloaded and extracted files will be stored.
        limit: Maximum number of pdf sources to download and extract.

    Raises:
        FileNotFoundError: If the source_links_file does not exist.

    Returns:
        str: Path to the output directory where the downloaded and extracted files are stored.
    """

    if source_links_file is None:
        source_links_file = os.path.join(
            os.path.dirname(__file__), "sources", "arxiv_urls.jsonl"
        )

    if not os.path.exists(source_links_file):
        raise FileNotFoundError(f"File '{source_links_file}' not found.")

    urls = pd.read_json(path_or_buf=source_links_file, lines=True)
    urls = urls[0].tolist()

    if limit:
        urls = urls[:limit]

    if output_dir is None:
        output_dir = os.path.join(
            os.path.dirname(__file__), "data", "raw", "arxiv_pdfs"
        )

    os.makedirs(output_dir, exist_ok=True)
    output_jsonl_dir = os.path.join(output_dir, "jsonl")
    os.makedirs(output_jsonl_dir, exist_ok=True)

    downloader = ArxivDownloader(output_dir)
    iterator = ArxivIterator()
    extractor = ArxivExtractor()

    output_format = {
        "text": str,
        "id": str,
        "file_extension": str,
        "file_type": str,
        "category": str,
        "line_count": int,
        "size_in_bytes": int,
        "path": str,
        "file_name": str,
    }

    dataset = download_and_extract(
        urls=urls,
        output_paths=[
            os.path.join(output_jsonl_dir, os.path.basename(url)) for url in urls
        ],
        downloader=downloader,
        iterator=iterator,
        extractor=extractor,
        output_format=output_format,
        keep_raw_download=True,
    )

    dataset.to_json(output_jsonl_dir, write_to_filename="file_name")
    return output_jsonl_dir
